"""
Run MAD Competition.

Classes to perform the synthesis of Maximum Differentiation Competition.
"""

import contextlib
import warnings
from collections import OrderedDict
from collections.abc import Callable
from typing import Any, Literal

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import torch
from pyrtools.tools.display import make_figure as pt_make_figure
from torch import Tensor
from tqdm.auto import tqdm

from ..tools import data, display, optim
from ..tools.convergence import loss_convergence
from ..tools.validate import validate_input, validate_metric
from .synthesis import OptimizedSynthesis


class MADCompetition(OptimizedSynthesis):
    r"""
    Synthesize a single maximally-differentiating image for two metrics.

    Following the basic idea in [1]_, this class synthesizes a
    maximally-differentiating image for two given metrics, based on a given
    image. We start by adding noise to this image and then iteratively
    adjusting its pixels so as to either minimize or maximize
    ``optimized_metric`` while holding the value of ``reference_metric`` constant.

    MADCompetiton accepts two metrics as its input. These should be callables
    that take two images and return a single number, and that number should be
    0 if and only if the two images are identical (thus, the larger the number,
    the more different the two images).

    Note that a full set of MAD Competition images consists of two pairs: a maximal and
    a minimal image for each metric. A single instantiation of ``MADCompetition`` will
    generate one of these four images.

    Parameters
    ----------
    image
        A tensor, this is the image we use as the reference point.
    optimized_metric
        The metric whose value you wish to minimize or maximize, which takes
        two tensors and returns a scalar.
    reference_metric
        The metric whose value you wish to keep fixed, which takes two tensors
        and returns a scalar.
    minmax
        Whether you wish to minimize or maximize ``optimized_metric``.
    metric_tradeoff_lambda
        Lambda to multiply by ``reference_metric`` loss and add to
        ``optimized_metric`` loss. If ``None``, we pick a value so the two
        initial losses are approximately equal in magnitude.
    range_penalty_lambda
        Lambda to multiply by range penalty and add to loss.
    allowed_range
        Range (inclusive) of allowed pixel values. Any values outside this
        range will be penalized.

    References
    ----------
    .. [1] Wang, Z., & Simoncelli, E. P. (2008). Maximum differentiation (MAD)
           competition: A methodology for comparing computational models of
           perceptual discriminability. Journal of Vision, 8(12), 1–13.
           https://dx.doi.org/10.1167/8.12.8
    """

    def __init__(
        self,
        image: Tensor,
        optimized_metric: torch.nn.Module | Callable[[Tensor, Tensor], Tensor],
        reference_metric: torch.nn.Module | Callable[[Tensor, Tensor], Tensor],
        minmax: Literal["min", "max"],
        metric_tradeoff_lambda: float | None = None,
        range_penalty_lambda: float = 0.1,
        allowed_range: tuple[float, float] = (0, 1),
    ):
        super().__init__(range_penalty_lambda, allowed_range)
        validate_input(image, allowed_range=allowed_range)
        validate_metric(
            optimized_metric,
            image_shape=image.shape,
            image_dtype=image.dtype,
            device=image.device,
        )
        validate_metric(
            reference_metric,
            image_shape=image.shape,
            image_dtype=image.dtype,
            device=image.device,
        )
        self._optimized_metric = optimized_metric
        self._reference_metric = reference_metric
        self._image = image.detach()
        self._image_shape = image.shape
        self._scheduler = None
        self._scheduler_step_arg = False
        self._optimized_metric_loss = []
        self._reference_metric_loss = []
        if minmax not in ["min", "max"]:
            raise ValueError(
                "synthesis_target must be one of {'min', 'max'}, but got "
                f"value {minmax} instead!"
            )
        self._mad_image = None
        self._initial_image = None
        self._reference_metric_target = None
        # If no metric_tradeoff_lambda is specified, pick one that gets them to
        # approximately the same magnitude
        if metric_tradeoff_lambda is None:
            other_image = torch.rand_like(image)
            optim_loss = optimized_metric(image, other_image)
            loss_ratio = optim_loss / reference_metric(image, other_image)
            metric_tradeoff_lambda = torch.pow(
                torch.as_tensor(10), torch.round(torch.log10(loss_ratio))
            ).item()
            warnings.warn(
                "Since metric_tradeoff_lamda was None, automatically set"
                f" to {metric_tradeoff_lambda} to roughly balance metrics."
            )
        self._metric_tradeoff_lambda = metric_tradeoff_lambda
        self._minmax = minmax
        self._store_progress = None
        self._saved_mad_image = []

    def setup(
        self,
        initial_noise: float | None = None,
        optimizer: torch.optim.Optimizer | None = None,
        optimizer_kwargs: dict | None = None,
        scheduler: torch.optim.lr_scheduler.LRScheduler | None = None,
        scheduler_kwargs: dict | None = None,
    ):
        """
        Initialize the MAD image, optimizer, and scheduler.

        Can only be called once. If ``load()`` has been called, ``initial_noise`` must
        be None.

        Parameters
        ----------
        initial_noise
            :attr:`mad_image` is initialized to ``self.image + initial_noise *
            torch.randn_like(self.image)``, so this gives the standard deviation of the
            Gaussian noise. If ``None``, we use a value of 0.1.
        optimizer
            The un-initialized optimizer object to use. If ``None``, we use Adam.
        optimizer_kwargs
            The keyword arguments to pass to the optimizer on initialization. If
            ``None``, we use ``{"lr": .01}`` and, if optimizer is ``None``,
            ``{"amsgrad": True}``.
        scheduler
            The un-initialized learning rate scheduler object to use. If ``None``, we
            don't use one.
        scheduler_kwargs
            The keyword arguments to pass to the scheduler on initialization.

        Raises
        ------
        ValueError
            If you try to set ``initial_noise`` after calling :func:`load`.
        ValueError
            If ``setup`` is called more than once or after :func:`synthesize`.

        Examples
        --------
        Set initial noise:

        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> mad = po.synth.MADCompetition(
        ...     img,
        ...     lambda x, y: 1 - po.metric.ssim(x, y),
        ...     po.metric.mse,
        ...     "min",
        ...     metric_tradeoff_lambda=0.1,
        ... )
        >>> mad.setup(1)
        >>> mad.synthesize(10)

        Set optimizer:

        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> mad = po.synth.MADCompetition(
        ...     img,
        ...     lambda x, y: 1 - po.metric.ssim(x, y),
        ...     po.metric.mse,
        ...     "min",
        ...     metric_tradeoff_lambda=0.1,
        ... )
        >>> mad.setup(optimizer=torch.optim.SGD, optimizer_kwargs={"lr": 0.01})
        >>> mad.synthesize(10)

        Use with save/load. Only the optimizer object is necessary, its kwargs and the
        initial noise are handled by load.

        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> mad = po.synth.MADCompetition(
        ...     img,
        ...     lambda x, y: 1 - po.metric.ssim(x, y),
        ...     po.metric.mse,
        ...     "min",
        ...     metric_tradeoff_lambda=0.1,
        ... )
        >>> mad.setup(1, optimizer=torch.optim.SGD, optimizer_kwargs={"lr": 0.01})
        >>> mad.synthesize(10)
        >>> mad.save("mad_setup.pt")
        >>> mad = po.synth.MADCompetition(
        ...     img,
        ...     lambda x, y: 1 - po.metric.ssim(x, y),
        ...     po.metric.mse,
        ...     "min",
        ...     metric_tradeoff_lambda=0.1,
        ... )
        >>> mad.load("mad_setup.pt")
        >>> mad.setup(optimizer=torch.optim.SGD)
        >>> mad.synthesize(10)
        """
        if self._mad_image is None:
            if initial_noise is None:
                initial_noise = 0.1
            mad_image = self.image + initial_noise * torch.randn_like(self.image)
            mad_image = mad_image.clamp(*self.allowed_range)
            self._initial_image = mad_image.clone()
            mad_image.requires_grad_()
            self._mad_image = mad_image
            self._reference_metric_target = self.reference_metric(
                self.image, self.mad_image
            ).item()
            self._reference_metric_loss.append(self._reference_metric_target)
            self._optimized_metric_loss.append(
                self.optimized_metric(self.image, self.mad_image).item()
            )
        else:
            if self._loaded:
                if initial_noise is not None:
                    raise ValueError("Cannot set initial_noise after calling load()!")
            else:
                raise ValueError(
                    "setup() can only be called once and must be called"
                    " before synthesize()!"
                )

        # initialize the optimizer
        self._initialize_optimizer(optimizer, self.mad_image, optimizer_kwargs)
        # and scheduler
        self._initialize_scheduler(scheduler, self.optimizer, scheduler_kwargs)
        # reset _loaded, if everything ran successfully
        self._loaded = False

    def synthesize(
        self,
        max_iter: int = 100,
        store_progress: bool | int = False,
        stop_criterion: float = 1e-4,
        stop_iters_to_check: int = 50,
    ):
        r"""
        Synthesize a MAD image.

        Update the pixels of :attr:`initial_image` to maximize or minimize
        (depending on the value of ``minmax``) the value of
        ``optimized_metric(image, mad_image)`` while keeping the value of
        ``reference_metric(image, mad_image)`` constant.

        We run this until either we reach ``max_iter`` or the change over the
        past ``stop_iters_to_check`` iterations is less than
        ``stop_criterion``, whichever comes first.

        Parameters
        ----------
        max_iter
            The maximum number of iterations to run before we end synthesis
            (unless we hit the stop criterion).
        store_progress
            Whether we should store the MAD image in progress during synthesis. If
            ``False``, we don't save anything. If True, we save every iteration. If an
            int, we save every ``store_progress`` iterations (note then that ``0`` is
            the same as ``False`` and ``1`` the same as ``True``).
        stop_criterion
            If the loss over the past ``stop_iters_to_check`` has changed
            less than ``stop_criterion``, we terminate synthesis.
        stop_iters_to_check
            How many iterations back to check in order to see if the
            loss has stopped decreasing (for ``stop_criterion``).

        Raises
        ------
        ValueError
            If we find a NaN during optimization.
        """
        # if setup hasn't been called manually, call it now.
        if self._mad_image is None or isinstance(self._scheduler, tuple):
            self.setup()
        self._current_loss = None

        # get ready to store progress
        self.store_progress = store_progress

        pbar = tqdm(range(max_iter))

        for _ in pbar:
            # update saved_* attrs. len(_losses) gives the total number of
            # iterations and will be correct across calls to `synthesize`
            self._store(len(self._losses))

            loss = self._optimizer_step(pbar)

            if not np.isfinite(loss):
                raise ValueError("Found a NaN in loss during optimization.")

            if self._check_convergence(stop_criterion, stop_iters_to_check):
                warnings.warn("Loss has converged, stopping synthesis")
                break

        # compute current loss, no need to compute gradient
        with torch.no_grad():
            self._current_loss = self.objective_function().item()

        pbar.close()

    def objective_function(
        self,
        mad_image: Tensor | None = None,
        image: Tensor | None = None,
    ) -> Tensor:
        r"""
        Compute the MADCompetition synthesis loss.

        This computes:

        .. math::

            t L_1(x, \hat{x}) &+ \lambda_1 [L_2(x, x+\epsilon) - L_2(x, \hat{x})]^2 \\
                              &+ \lambda_2 \mathcal{B}(\hat{x})

        where :math:`t` is 1 if :attr:`minmax` is ``'min'`` and -1 if it's ``'max'``,
        :math:`L_1` is :attr:`optimized_metric`, :math:`L_2` is
        :attr:`reference_metric`, :math:`x` is :attr:`image`, :math:`\hat{x}` is
        :attr:`mad_image`, :math:`\epsilon` is the initial noise, :math:`\mathcal{B}` is
        the quadratic bound penalty, :math:`\lambda_1` is :attr:`metric_tradeoff_lambda`
        and :math:`\lambda_2` is :attr:`range_penalty_lambda`.

        Parameters
        ----------
        mad_image
            Proposed ``mad_image``, :math:`\hat{x}` in the above equation. If
            ``None``, use ``self.mad_image``.
        image
            Proposed ``image``, :math:`x` in the above equation. If
            ``None``, use ``self.image``.

        Returns
        -------
        loss
            1-element tensor containing the loss on this step.
        """
        if image is None:
            image = self.image
        if mad_image is None:
            mad_image = self.mad_image
            # if this is empty, then self.mad_image hasn't been initialized
            if mad_image.numel() == 0:
                return torch.empty(0)
        synth_target = {"min": 1, "max": -1}[self.minmax]
        synthesis_loss = self.optimized_metric(image, mad_image)
        fixed_loss = (
            self._reference_metric_target - self.reference_metric(image, mad_image)
        ).pow(2)
        range_penalty = optim.penalize_range(mad_image, self.allowed_range)
        return (
            synth_target * synthesis_loss
            + self.metric_tradeoff_lambda * fixed_loss
            + self.range_penalty_lambda * range_penalty
        )

    def get_progress(
        self,
        iteration: int,
        iteration_selection: Literal["floor", "ceiling", "round"] = "round",
    ) -> dict:
        """
        Return dictionary summarizing synthesis progress at ``iteration``.

        This returns a dictionary containing info from :attr:`losses`,
        :attr:`pixel_change_norm`, :attr:`gradient_norm`, and
        :attr:`saved_mad_image` corresponding to ``iteration``. If synthesis was
        run with ``store_progress=False`` (and so we did not cache anything in
        :attr:`saved_mad_image`), then that key will be missing. If synthesis was
        run with ``store_progress>1``, we will grab the corresponding tensor
        from :attr:`saved_mad_image`, with behavior determined by
        ``iteration_selection``.

        The returned dictionary will additionally contain the keys:

        - ``"iteration"``: the (0-indexed positive) synthesis iteration that the
          values for :attr:`losses`, :attr:`pixel_change_norm`, and
          :attr:`gradient_norm` come from.

        - If ``self.store_progress``, ``"store_progress_iteration"``: the (0-indexed
          positive) synthesis iteration that the value for :attr:`saved_mad_image` comes
          from.

        Note that for the most recent iteration (``iteration=-1`` or ``iteration=None``
        or ``iteration==len(self.losses)-1``), we do not have values for
        :attr:`pixel_change_norm` or :attr:`gradient_norm`, since in this case we are
        showing the loss and value for the current MAD image.

        Parameters
        ----------
        iteration
            Synthesis iteration to summarize. If ``None``, grab the most recent.
            Negative values are allowed.
        iteration_selection

            How to select the relevant iteration from :attr:`saved_mad_image`
            when the request iteration wasn't stored.

            When synthesis was run with ``store_progress=n`` (where ``n>1``),
            MAD images are only saved every ``n`` iterations. If you request an
            iteration where a MAD image wasn't saved, this determines which available
            iteration is used instead:

            * ``"floor"``: use the closest saved iteration **before** the
              requested one.

            * ``"ceiling"``: use the closest saved iteration **after** the
              requested one.

            * ``"round"``: use the closest saved iteration.

        Returns
        -------
        progress_info
            Dictionary summarizing synthesis progress.

        Raises
        ------
        IndexError
            If ``iteration`` takes an illegal value.

        Warns
        -----
        UserWarning
            If the iteration used for ``saved_mad_image`` is not the same as the
            argument ``iteration`` (because e.g., you set ``iteration=3`` but
            ``self.store_progress=2``).
        """
        return super().get_progress(
            iteration,
            iteration_selection,
            ["reference_metric_loss", "optimized_metric_loss"],
            store_progress_attributes=["saved_mad_image"],
        )

    def _optimizer_step(self, pbar: tqdm) -> Tensor:
        r"""
        Compute and propagate gradients, then step optimizer to update mad_image.

        Parameters
        ----------
        pbar
            A tqdm progress-bar, which we update with a postfix
            describing the current loss, gradient norm, and learning
            rate (it already tells us which iteration and the time
            elapsed).

        Returns
        -------
        loss
            1-element tensor containing the loss on this step.
        """  # numpydoc ignore=ES01
        last_iter_mad_image = self.mad_image.clone()
        loss = self.optimizer.step(self._closure)
        self._losses.append(loss)
        grad_norm = torch.linalg.vector_norm(self.mad_image.grad.data, ord=2, dim=None)
        self._gradient_norm.append(grad_norm.item())

        fm = self.reference_metric(self.image, self.mad_image)
        self._reference_metric_loss.append(fm.item())
        sm = self.optimized_metric(self.image, self.mad_image)
        self._optimized_metric_loss.append(sm.item())

        # optionally step the scheduler, passing loss if needed
        if self.scheduler is not None:
            if self._scheduler_step_arg:
                self.scheduler.step(loss)
            else:
                self.scheduler.step()

        pixel_change_norm = torch.linalg.vector_norm(
            self.mad_image - last_iter_mad_image, ord=2, dim=None
        )
        self._pixel_change_norm.append(pixel_change_norm.item())

        # add extra info here if you want it to show up in progress bar
        pbar.set_postfix(
            OrderedDict(
                loss=f"{loss:.04e}",
                learning_rate=self.optimizer.param_groups[0]["lr"],
                gradient_norm=f"{grad_norm.item():.04e}",
                pixel_change_norm=f"{pixel_change_norm.item():.04e}",
                reference_metric=f"{fm.item():.04e}",
                optimized_metric=f"{sm.item():.04e}",
            )
        )
        return loss

    def _check_convergence(
        self, stop_criterion: float, stop_iters_to_check: int
    ) -> bool:
        r"""
        Check whether the loss has stabilized and, if so, return True.

        Uses :func:`loss_convergence`.

        Parameters
        ----------
        stop_criterion
            If the loss over the past ``stop_iters_to_check`` has changed
            less than ``stop_criterion``, we terminate synthesis.
        stop_iters_to_check
            How many iterations back to check in order to see if the
            loss has stopped decreasing (for ``stop_criterion``).

        Returns
        -------
        loss_stabilized
            Whether the loss has stabilized or not.
        """
        return loss_convergence(self, stop_criterion, stop_iters_to_check)

    def _store(self, i: int) -> bool:
        """
        Store mad_image and model response, if appropriate.

        If it's the right iteration, we update :attr:`saved_mad_image`.

        Parameters
        ----------
        i
            The current iteration.

        Returns
        -------
        stored
            True if we stored this iteration, False if not.
        """
        if self.store_progress and (i % self.store_progress == 0):
            # want these to always be on cpu, to reduce memory use for GPUs
            self._saved_mad_image.append(self.mad_image.clone().to("cpu"))
            stored = True
        else:
            stored = False
        return stored

    def save(self, file_path: str):
        r"""
        Save all relevant variables in .pt file.

        Note that if ``store_progress`` is True, this will probably be very
        large.

        See :func:`load` docstring for an example of use.

        Parameters
        ----------
        file_path
            The path to save the MADCompetition object to.
        """
        save_io_attrs = [
            ("_optimized_metric", ("_image", "_mad_image")),
            ("_reference_metric", ("_image", "_mad_image")),
        ]
        save_state_dict_attrs = ["_optimizer", "_scheduler"]
        super().save(file_path, save_io_attrs, save_state_dict_attrs)

    def to(self, *args: Any, **kwargs: Any):
        r"""
        Move and/or casts the parameters and buffers.

        This can be called as

        .. function:: to(device=None, dtype=None, non_blocking=False)

        .. function:: to(dtype, non_blocking=False)

        .. function:: to(tensor, non_blocking=False)

        Its signature is similar to :meth:`torch.Tensor.to`, but only accepts
        floating point desired :attr:`dtype` s. In addition, this method will
        only cast the floating point parameters and buffers to :attr:`dtype`
        (if given). The integral parameters and buffers will be moved
        :attr:`device`, if that is given, but with dtypes unchanged. When
        :attr:`non_blocking` is set, it tries to convert/move asynchronously
        with respect to the host if possible, e.g., moving CPU Tensors with
        pinned memory to CUDA devices.

        See :meth:`torch.nn.Module.to` for examples.

        .. note::
            This method modifies the module in-place.

        Parameters
        ----------
        device : torch.device
            The desired device of the parameters and buffers in this module.
        dtype : torch.dtype
            The desired floating point type of the floating point parameters and
            buffers in this module.
        tensor : torch.Tensor
            Tensor whose dtype and device are the desired dtype and device for
            all parameters and buffers in this module.
        """  # numpydoc ignore=PR01,PR02
        attrs = ["_initial_image", "_image", "_mad_image", "_saved_mad_image"]
        super().to(*args, attrs=attrs, **kwargs)
        # if the metrics are Modules, then we should pass them as well. If
        # they're functions then nothing needs to be done.
        with contextlib.suppress(AttributeError):
            self.reference_metric.to(*args, **kwargs)
        with contextlib.suppress(AttributeError):
            self.optimized_metric.to(*args, **kwargs)

    def load(
        self,
        file_path: str,
        map_location: str | None = None,
        tensor_equality_atol: float = 1e-8,
        tensor_equality_rtol: float = 1e-5,
        **pickle_load_args: Any,
    ):
        r"""
        Load all relevant stuff from a .pt file.

        This must be called by a ``MADCompetition`` object initialized just like the
        saved object.

        Note this operates in place and so doesn't return anything.

        .. versionchanged:: 1.2
           load behavior changed in a backwards-incompatible manner in order to
           compatible with breaking changes in torch 2.6.

        Parameters
        ----------
        file_path
            The path to load the synthesis object from.
        map_location
            Argument to pass to ``torch.load`` as ``map_location``. If you save
            stuff that was being run on a GPU and are loading onto a
            CPU, you'll need this to make sure everything lines up
            properly. This should be structured like the str you would
            pass to :class:`torch.device`.
        tensor_equality_atol
            Absolute tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        tensor_equality_rtol
            Relative tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        **pickle_load_args
            Any additional kwargs will be added to ``pickle_module.load`` via
            :func:`torch.load`, see that function's docstring for details.

        Raises
        ------
        ValueError
            If :func:`setup` or :func:`synthesize` has been called before this call
            to ``load``.
        ValueError
            If the object saved at ``file_path`` is not a ``MADCompetition`` object.
        ValueError
            If the saved and loading ``MADCompetition`` objects have a different value
            for any of :attr:`image`, :attr:`range_penalty_lambda`,
            :attr:`allowed_range`, :attr:`metric_tradeoff_lambda`, or :attr:`minimax`.
        ValueError
            If the behavior of :attr:`optimized_metric` or :attr:`reference_metric` is
            different between the saved and loading objects.

        Warns
        -----
        UserWarning
            If :func:`setup` will need to be called after ``load``, to finish
            initializing :attr:`optimizer` or :attr:`scheduler`.

        See Also
        --------
        :func:`~plenoptic.tools.io.examine_saved_synthesis`
            Examine metadata from saved object: pytorch and plenoptic versions, name of
            the synthesis object, shapes of tensors, etc.

        Examples
        --------
        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> def ds_ssim(x, y):
        ...     return 1 - po.metric.ssim(x, y)
        >>> mad = po.synth.MADCompetition(
        ...     img, po.metric.mse, ds_ssim, "min", metric_tradeoff_lambda=10
        ... )
        >>> mad.synthesize(max_iter=5, store_progress=True)
        >>> mad.save("mad.pt")
        >>> mad_copy = po.synth.MADCompetition(
        ...     img, po.metric.mse, ds_ssim, "min", metric_tradeoff_lambda=10
        ... )
        >>> mad_copy.load("mad.pt")
        """
        check_attributes = [
            "_image",
            "_metric_tradeoff_lambda",
            "_range_penalty_lambda",
            "_allowed_range",
            "_minmax",
        ]
        check_io_attrs = [
            ("_optimized_metric", ("_image", "_mad_image")),
            ("_reference_metric", ("_image", "_mad_image")),
        ]
        super().load(
            file_path,
            "losses",
            map_location=map_location,
            check_attributes=check_attributes,
            check_io_attributes=check_io_attrs,
            state_dict_attributes=["_optimizer", "_scheduler"],
            tensor_equality_atol=tensor_equality_atol,
            tensor_equality_rtol=tensor_equality_rtol,
            **pickle_load_args,
        )
        # make this require a grad again
        self.mad_image.requires_grad_()
        # these are always supposed to be on cpu, but may get copied over to
        # gpu on load (which can cause problems when resuming synthesis), so
        # fix that.
        if len(self._saved_mad_image) and self._saved_mad_image[0].device.type != "cpu":
            self._saved_mad_image = [mad.to("cpu") for mad in self._saved_mad_image]

    @property
    def mad_image(self) -> Tensor:
        """Maximally-differentiating image, the parameter we are optimizing."""
        # numpydoc ignore=RT01,ES01
        if self._mad_image is None:
            return torch.empty(0)
        return self._mad_image

    @property
    def optimized_metric(self) -> torch.nn.Module | Callable[[Tensor, Tensor], Tensor]:
        """The metric whose value we are minimizing or maximizing."""
        # numpydoc ignore=RT01,ES01
        return self._optimized_metric

    @property
    def reference_metric(self) -> torch.nn.Module | Callable[[Tensor, Tensor], Tensor]:
        """The metric whose value we are keeping constant."""
        # numpydoc ignore=RT01,ES01
        return self._reference_metric

    @property
    def image(self) -> Tensor:
        """The reference image for this MAD Competition."""
        # numpydoc ignore=RT01,ES01
        return self._image

    @property
    def initial_image(self) -> Tensor:
        """
        Initial image for MAD Competition.

        This is the image whose distance to ``image``, the reference, we are
        maximizing/minimizing for ``optimized_metric``, while keeping constant for
        ``reference_metric``.
        """
        # numpydoc ignore=RT01
        return self._initial_image

    @property
    def reference_metric_loss(self) -> Tensor:
        """
        :attr:`reference_metric` loss over iterations.

        That is, the value of ``reference_metric(image, mad_image)``. Ideally, this is
        equal to ``reference_metric(image, initial_image)``.

        This tensor always lives on the CPU, regardless of the device of the
        ``MADCompetition`` object.
        """
        # numpydoc ignore=RT01
        return torch.as_tensor(self._reference_metric_loss)

    @property
    def optimized_metric_loss(self) -> Tensor:
        """
        :attr:`optimized_metric` loss over iterations.

        That is, the value of ``optimized_metric(image, mad_image)``. Ideally, this is
        either very different from ``optimized_metric(image, initial_image)``.

        This tensor always lives on the CPU, regardless of the device of the
        ``MADCompetition`` object.
        """
        # numpydoc ignore=RT01
        return torch.as_tensor(self._optimized_metric_loss)

    @property
    def metric_tradeoff_lambda(self) -> float:
        """Tradeoff between the two metrics in synthesis loss."""
        # numpydoc ignore=RT01,ES01
        return self._metric_tradeoff_lambda

    @property
    def minmax(self) -> Literal["min", "max"]:
        """Whether we are minimizing or maximizing :attr:`optimized_metric`."""
        # numpydoc ignore=RT01,ES01
        return self._minmax

    @property
    def saved_mad_image(self) -> Tensor:
        """
        :attr:`mad_image`, cached over time for later examination.

        How often the MAD image is cached is determined by the ``store_progress``
        argument to the :func:`synthesize` function.

        The last entry will always be the current :attr:`mad_image`.

        If ``store_progress==1``, then this corresponds directly to :attr:`losses`:
        ``losses[i]`` is the error for ``saved_mad_image[i]``

        This tensor always lives on the CPU, regardless of the device of the
        ``MADCompetition`` object.
        """  # numpydoc ignore=RT01
        if self._mad_image is None:
            return torch.empty(0)
        else:
            # for memory purposes, always on CPU
            return torch.stack([*self._saved_mad_image, self.mad_image.to("cpu")])


def plot_loss(
    mad: MADCompetition,
    iteration: int | None = None,
    axes: list[mpl.axes.Axes] | mpl.axes.Axes | None = None,
    **kwargs: Any,
) -> mpl.axes.Axes:
    """
    Plot metric losses.

    Plots ``mad.optimized_metric_loss`` and ``mad.reference_metric_loss`` on two
    separate axes, over all iterations. Also plots a red dot at ``iteration``,
    to highlight the loss there. If ``iteration=None``, then the dot will be at
    the final iteration.

    Parameters
    ----------
    mad
        MADCompetition object whose loss we want to plot.
    iteration
        Which iteration to display. If ``None``, we show the most recent one.
        Negative values are also allowed.
    axes
        Pre-existing axes for plot. If a list of axes, must be the two axes to
        use for this plot. If a single axis, we'll split it in half
        horizontally. If ``None``, we call :func:`matplotlib.pyplot.gca()`.
    **kwargs
        Passed to :func:`matplotlib.pyplot.semilogy`.

    Returns
    -------
    axes :
        The matplotlib axes containing the plot.

    Raises
    ------
    IndexError
        If ``iteration`` takes an illegal value.

    Notes
    -----
    We plot ``abs(mad.losses)`` because if we're maximizing the synthesis
    metric, we minimized its negative. By plotting the absolute value, we get
    them all on the same scale.
    """
    # this warning is not relevant for this plotting function
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="loss iteration and iteration for")
        progress = mad.get_progress(iteration)

    if axes is None:
        axes = plt.gca()
    if not hasattr(axes, "__iter__"):
        axes = display.clean_up_axes(
            axes, False, ["top", "right", "bottom", "left"], ["x", "y"]
        )
        gs = axes.get_subplotspec().subgridspec(1, 2)
        fig = axes.figure
        axes = [fig.add_subplot(gs[0, 0]), fig.add_subplot(gs[0, 1])]
    losses = [mad.reference_metric_loss, mad.optimized_metric_loss]
    names = ["reference_metric_loss", "optimized_metric_loss"]
    for ax, loss, name in zip(axes, losses, names):
        ax.plot(loss, **kwargs)
        ax.scatter(progress["iteration"], progress[name], c="r")
        ax.set(xlabel="Synthesis iteration", ylabel=name.capitalize().replace("_", " "))
    return ax


def display_mad_image(
    mad: MADCompetition,
    batch_idx: int = 0,
    channel_idx: int | None = None,
    zoom: float | None = None,
    iteration: int | None = None,
    ax: mpl.axes.Axes | None = None,
    title: str | None = None,
    **kwargs: Any,
) -> mpl.axes.Axes:
    """
    Display MAD image.

    We use :func:`~plenoptic.tools.display.imshow` to display the synthesized image and
    attempt to automatically find the most reasonable zoom value. You can override this
    value using the zoom arg, but remember that :func:`~plenoptic.tools.display.imshow`
    is opinionated about the size of the resulting image and will throw an Exception if
    the axis created is not big enough for the selected zoom.

    Parameters
    ----------
    mad
        MADCompetition object whose MAD image we want to display.
    batch_idx
        Which index to take from the batch dimension.
    channel_idx
        Which index to take from the channel dimension. If ``None``, we assume
        image is RGB(A) and show all channels.
    zoom
        How much to zoom in / enlarge the synthesized image, the ratio
        of display pixels to image pixels. If ``None``, we
        attempt to find the best value ourselves.
    iteration
        Which iteration to display. If ``None``, we show the most recent one.
        Negative values are also allowed. If ``iteration!=None`` and
        ``mad.store_progress>1`` (that is, the MAD image was not cached on every
        iteration), then we show the cached MAD image from the nearest iteration.
    ax
        Pre-existing axes for plot. If ``None``, we call :func:`matplotlib.pyplot.gca`.
    title
        Title to add to axis. If ``None``, we use ``"MAD Image [iteration={iter}]"``,
        where ``iter`` gives the iteration corresponding to the displayed image.
    **kwargs
        Passed to :func:`~plenoptic.tools.display.imshow`.

    Returns
    -------
    ax :
        The matplotlib axes containing the plot.

    Raises
    ------
    ValueError
        If ``batch_idx`` is not an int.
    IndexError
        If ``iteration`` takes an illegal value.

    Warns
    -----
    UserWarning
        If the iteration used for ``saved_mad_image`` is not the same as the argument
        ``iteration`` (because e.g., you set ``iteration=3`` but
        ``mad.store_progress=2``).
    """
    progress = mad.get_progress(iteration)
    try:
        image = progress["saved_mad_image"]
        iter = progress["store_progress_iteration"]
    except KeyError:
        if iteration is not None:
            raise IndexError("When mad.store_progress=False, iteration must be None!")
        image = mad.mad_image
        # losses will always have one extra value, the current loss.
        iter = len(mad.losses) - 1

    if batch_idx is None:
        raise ValueError("batch_idx must be an integer!")
    # we're only plotting one image here, so if the user wants multiple
    # channels, they must be RGB
    as_rgb = bool(channel_idx is None and image.shape[1] > 1)
    if ax is None:
        ax = plt.gca()
    if title is None:
        title = f"MAD Image [iteration={iter}]"
    display.imshow(
        image,
        ax=ax,
        title=title,
        zoom=zoom,
        batch_idx=batch_idx,
        channel_idx=channel_idx,
        as_rgb=as_rgb,
        **kwargs,
    )
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    return ax


def plot_pixel_values(
    mad: MADCompetition,
    batch_idx: int = 0,
    channel_idx: int | None = None,
    iteration: int | None = None,
    ylim: tuple[float] | Literal[False] = False,
    ax: mpl.axes.Axes | None = None,
    **kwargs: Any,
) -> mpl.axes.Axes:
    r"""
    Plot histogram of pixel values of reference and MAD images.

    As a way to check the distributions of pixel intensities and see
    if there's any values outside the allowed range.

    Parameters
    ----------
    mad
        MADCompetition object with the images whose pixel values we want to compare.
    batch_idx
        Which index to take from the batch dimension.
    channel_idx
        Which index to take from the channel dimension. If ``None``, we use all
        channels (assumed use-case is RGB(A) images).
    iteration
        Which iteration to display. If ``None``, we show the most recent one.
        Negative values are also allowed. If ``iteration!=None`` and
        ``mad.store_progress>1`` (that is, the MAD image was not cached on every
        iteration), then we use the cached MAD image from the nearest iteration.
    ylim
        If tuple, the ylimit to set for this axis. If ``False``, we leave
        it untouched.
    ax
        Pre-existing axes for plot. If ``None``, we call
        :func:`matplotlib.pyplot.gca()`.
    **kwargs
        Passed to :func:`matplotlib.pyplot.hist`.

    Returns
    -------
    ax :
        Creates axes.

    Raises
    ------
    IndexError
        If ``iteration`` takes an illegal value.

    Warns
    -----
    UserWarning
        If the iteration used for ``saved_mad_image`` is not the same as the argument
        ``iteration`` (because e.g., you set ``iteration=3`` but
        ``mad.store_progress=2``).
    """

    def _freedman_diaconis_bins(a: np.ndarray) -> int:
        """
        Calculate number of hist bins using Freedman-Diaconis rule.

        Copied from seaborn.

        Parameters
        ----------
        a
            The array to histogram.

        Returns
        -------
        n_bins
            Number of bins to use for histogram.
        """
        # From https://stats.stackexchange.com/questions/798/
        a = np.asarray(a)
        iqr = np.diff(np.percentile(a, [0.25, 0.75]))[0]
        if len(a) < 2:
            return 1
        h = 2 * iqr / (len(a) ** (1 / 3))
        # fall back to sqrt(a) bins if iqr is 0
        if h == 0:
            return int(np.sqrt(a.size))
        else:
            return int(np.ceil((a.max() - a.min()) / h))

    kwargs.setdefault("alpha", 0.4)
    progress = mad.get_progress(iteration)
    try:
        mad_image = progress["saved_mad_image"]
        iter = progress["store_progress_iteration"]
    except KeyError:
        if iteration is not None:
            raise IndexError("When mad.store_progress=False, iteration must be None!")
        mad_image = mad.mad_image
        # losses will always have one extra value, the current loss.
        iter = len(mad.losses) - 1
    image = mad.image[batch_idx]
    if channel_idx is not None:
        image = image[channel_idx]
        mad_image = mad_image[channel_idx]
    image = data.to_numpy(image).flatten()
    mad_image = data.to_numpy(mad_image).flatten()

    if ax is None:
        ax = plt.gca()
    ax.hist(
        mad_image,
        bins=min(_freedman_diaconis_bins(image), 50),
        label=f"MAD image [iteration={iter}]",
        **kwargs,
    )
    ax.hist(
        image,
        bins=min(_freedman_diaconis_bins(image), 50),
        label="Reference image",
        **kwargs,
    )
    ax.legend()
    if ylim:
        ax.set_ylim(ylim)
    ax.set_title("Histogram of pixel values")
    return ax


def _check_included_plots(to_check: list[str] | dict[str, int], to_check_name: str):
    """
    Check whether the user wanted us to create plots that we can't.

    Helper function for :func:`plot_synthesis_status` and :func:`animate`.

    Raises a ``ValueError`` if ``to_check`` contains any values that are not allowed.

    Parameters
    ----------
    to_check
        The variable to check. We ensure that it doesn't contain any extra (not
        allowed) values. If a list, we check its contents. If a dict, we check
        its keys.
    to_check_name
        Name of the ``to_check`` variable, used in the error message.

    Raises
    ------
    ValueError
        If ``to_check`` takes an illegal value.
    """
    allowed_vals = [
        "display_mad_image",
        "plot_loss",
        "plot_pixel_values",
        "misc",
    ]
    try:
        vals = to_check.keys()
    except AttributeError:
        vals = to_check
    not_allowed = [v for v in vals if v not in allowed_vals]
    if not_allowed:
        raise ValueError(
            f"{to_check_name} contained value(s) {not_allowed}! "
            f"Only {allowed_vals} are permissible!"
        )


def _setup_synthesis_fig(
    fig: mpl.figure.Figure | None = None,
    axes_idx: dict[str, int] = {},
    figsize: tuple[float] | None = None,
    included_plots: list[str] = [
        "display_mad_image",
        "plot_loss",
        "plot_pixel_values",
    ],
    display_mad_image_width: float = 1,
    plot_loss_width: float = 2,
    plot_pixel_values_width: float = 1,
) -> tuple[mpl.figure.Figure, list[mpl.axes.Axes], dict[str, int]]:
    """
    Set up figure for :func:`plot_synthesis_status`.

    Creates figure with enough axes for the all the plots you want. Will
    also create index in ``axes_idx`` for them if you haven't done so already.

    If ``fig=None``, all axes will be on the same row and have the same width. If
    you want them to be on different rows, will need to initialize ``fig`` yourself
    and pass that in. For changing width, change the corresponding ``*_width`` arg,
    which gives width relative to other axes. So if you want the axis for the
    loss plot to be three times as wide as the others, set ``loss_width=3``.

    Parameters
    ----------
    fig
        The figure to plot on or ``None``. If ``None``, we create a new figure.
    axes_idx
        Dictionary specifying which axes contains which type of plot, allows for more
        fine-grained control of the resulting figure. Probably only helpful if fig is
        also defined. Possible keys: ``"loss"``, ``"pixel_values"``, ``"misc"``. Values
        should all be ints. If you tell this function to create a plot that doesn't have
        a corresponding key, we find the lowest int that is not already in the dict, so
        if you have axes that you want unchanged, place their idx in ``"misc"``.
    figsize
        The size of the figure to create. It may take a little bit of
        playing around to find a reasonable value. If ``None``, we attempt to
        make our best guess, aiming to have relative width=1 correspond to 5.
    included_plots
        Which plots to include. Must be some subset of ``'display_mad_image',
        'plot_loss', 'plot_pixel_values'``.
    display_mad_image_width
        Relative width of the axis for the synthesized image.
    plot_loss_width
        Relative width of the axis for loss plot.
    plot_pixel_values_width
        Relative width of the axis for image pixel intensities histograms.

    Returns
    -------
    fig
        The figure to plot on.
    axes
        List or array of axes contained in fig.
    axes_idx
        Dictionary identifying the idx for each plot type.
    """
    n_subplots = 0
    axes_idx = axes_idx.copy()
    width_ratios = []
    if "display_mad_image" in included_plots:
        n_subplots += 1
        width_ratios.append(display_mad_image_width)
        if "display_mad_image" not in axes_idx:
            axes_idx["display_mad_image"] = data._find_min_int(axes_idx.values())
    if "plot_loss" in included_plots:
        n_subplots += 1
        width_ratios.append(plot_loss_width)
        if "plot_loss" not in axes_idx:
            axes_idx["plot_loss"] = data._find_min_int(axes_idx.values())
    if "plot_pixel_values" in included_plots:
        n_subplots += 1
        width_ratios.append(plot_pixel_values_width)
        if "plot_pixel_values" not in axes_idx:
            axes_idx["plot_pixel_values"] = data._find_min_int(axes_idx.values())
    if fig is None:
        width_ratios = np.array(width_ratios)
        if figsize is None:
            # we want (5, 5) for each subplot, with a bit of room between
            # each subplot
            figsize = ((width_ratios * 5).sum() + width_ratios.sum() - 1, 5)
        width_ratios = width_ratios / width_ratios.sum()
        fig, axes = plt.subplots(
            1,
            n_subplots,
            figsize=figsize,
            gridspec_kw={"width_ratios": width_ratios},
        )
        if n_subplots == 1:
            axes = [axes]
    else:
        axes = fig.axes
    # make sure misc contains all the empty axes
    misc_axes = axes_idx.get("misc", [])
    if not hasattr(misc_axes, "__iter__"):
        misc_axes = [misc_axes]
    all_axes = []
    for i in axes_idx.values():
        # so if it's a list of ints
        if hasattr(i, "__iter__"):
            all_axes.extend(i)
        else:
            all_axes.append(i)
    misc_axes += [i for i, _ in enumerate(fig.axes) if i not in all_axes]
    axes_idx["misc"] = misc_axes
    return fig, axes, axes_idx


def plot_synthesis_status(
    mad: MADCompetition,
    batch_idx: int = 0,
    channel_idx: int | None = None,
    iteration: int | None = None,
    vrange: tuple[float] | str = "indep1",
    zoom: float | None = None,
    fig: mpl.figure.Figure | None = None,
    axes_idx: dict[str, int] = {},
    figsize: tuple[float] | None = None,
    included_plots: list[str] = [
        "display_mad_image",
        "plot_loss",
        "plot_pixel_values",
    ],
    width_ratios: dict[str, float] = {},
) -> tuple[mpl.figure.Figure, dict[str, int]]:
    r"""
    Make a plot showing synthesis status.

    We create several subplots to analyze this. The plots to include are
    specified by including their name in the ``included_plots`` list. All plots
    can be created separately using the method with the same name.

    Parameters
    ----------
    mad
        MADCompetition object whose status we want to plot.
    batch_idx
        Which index to take from the batch dimension.
    channel_idx
        Which index to take from the channel dimension. If ``None``, we use all
        channels (assumed use-case is RGB(A) image).
    iteration
        Which iteration to display. If ``None``, we show the most recent one.
        Negative values are also allowed. If ``iteration!=None`` and
        ``mad.store_progress>1`` (that is, the MAD image was not cached on every
        iteration), then we use the cached MAD image from the nearest iteration.
    vrange
        The vrange option to pass to ``display_mad_image()``. See
        docstring of :func:`~plenoptic.tools.display.imshow` for possible values.
    zoom
        How much to zoom in / enlarge the synthesized image, the ratio
        of display pixels to image pixels. If ``None``, we
        attempt to find the best value ourselves.
    fig
        If ``None``, we create a new figure. otherwise we assume this is
        an empty figure that has the appropriate size and number of
        subplots.
    axes_idx
        Dictionary specifying which axes contains which type of plot, allows
        for more fine-grained control of the resulting figure. Probably only
        helpful if fig is also defined. Possible keys: ``'mad_image',
        'loss', 'pixel_values', 'misc'``. Values should all be ints. If you
        tell this function to create a plot that doesn't have a corresponding
        key, we find the lowest int that is not already in the dict, so if you
        have axes that you want unchanged, place their idx in ``'misc'``.
    figsize
        The size of the figure to create. It may take a little bit of
        playing around to find a reasonable value. If ``None``, we attempt to
        make our best guess, aiming to have each axis be of size ``(5, 5)``.
    included_plots
        Which plots to include. Must be some subset of ``'display_mad_image',
        'plot_loss', 'plot_pixel_values'``.
    width_ratios
        If ``width_ratios`` is an empty dictionary, ``plot_loss`` will have
        double the width of the other plots. To change that, specify their
        relative widths using the keys: ['display_mad_image', 'plot_loss',
        'plot_pixel_values'] and floats specifying their relative width.

    Returns
    -------
    fig
        The figure containing this plot.
    axes_idx
        Dictionary giving index of each plot.

    Raises
    ------
    ValueError
        If ``mad.mad_image`` object is not 3d or 4d.
    ValueError
        If the ``iteration is not None`` and the given ``mad`` object was run
        with ``store_progress=False``.

    Warns
    -----
    UserWarning
        If the iteration used for ``saved_mad_image`` is not the same as the argument
        ``iteration`` (because e.g., you set ``iteration=3`` but
        ``mad.store_progress=2``).
    """
    if iteration is not None and not mad.store_progress:
        raise ValueError(
            "synthesis() was run with store_progress=False, "
            "cannot specify which iteration to plot (only"
            " last one, with iteration=None)"
        )
    if mad.mad_image.ndim not in [3, 4]:
        raise ValueError(
            "plot_synthesis_status() expects 3 or 4d data;"
            "unexpected behavior will result otherwise!"
        )
    _check_included_plots(included_plots, "included_plots")
    _check_included_plots(width_ratios, "width_ratios")
    _check_included_plots(axes_idx, "axes_idx")
    width_ratios = {f"{k}_width": v for k, v in width_ratios.items()}
    fig, axes, axes_idx = _setup_synthesis_fig(
        fig, axes_idx, figsize, included_plots, **width_ratios
    )

    if "display_mad_image" in included_plots:
        display_mad_image(
            mad,
            batch_idx=batch_idx,
            channel_idx=channel_idx,
            iteration=iteration,
            ax=axes[axes_idx["display_mad_image"]],
            zoom=zoom,
            vrange=vrange,
        )
    if "plot_loss" in included_plots:
        plot_loss(mad, iteration=iteration, axes=axes[axes_idx["plot_loss"]])
        # this function creates a single axis for loss, which plot_loss then
        # split into two. this makes sure the right two axes are present in the
        # dict
        all_axes = []
        for i in axes_idx.values():
            # so if it's a list of ints
            if hasattr(i, "__iter__"):
                all_axes.extend(i)
            else:
                all_axes.append(i)
        new_axes = [i for i, _ in enumerate(fig.axes) if i not in all_axes]
        axes_idx["plot_loss"] = new_axes
    if "plot_pixel_values" in included_plots:
        plot_pixel_values(
            mad,
            batch_idx=batch_idx,
            channel_idx=channel_idx,
            iteration=iteration,
            ax=axes[axes_idx["plot_pixel_values"]],
        )
    return fig, axes_idx


def animate(
    mad: MADCompetition,
    framerate: int = 10,
    batch_idx: int = 0,
    channel_idx: int | None = None,
    zoom: float | None = None,
    fig: mpl.figure.Figure | None = None,
    axes_idx: dict[str, int] = {},
    figsize: tuple[float] | None = None,
    included_plots: list[str] = [
        "display_mad_image",
        "plot_loss",
        "plot_pixel_values",
    ],
    width_ratios: dict[str, float] = {},
) -> mpl.animation.FuncAnimation:
    r"""
    Animate synthesis progress.

    This is essentially the figure produced by
    ``mad.plot_synthesis_status`` animated over time, for each stored
    iteration.

    This functions returns a matplotlib FuncAnimation object. See our
    documentation (e.g., :ref:`quickstart-nb`) for examples on how to view it in
    a Jupyter notebook. In order to save, use ``anim.save(filename)``. In either
    case, this can take a while and you'll need the appropriate writer installed
    and on your path, e.g., ffmpeg, imagemagick, etc). See
    :doc:`matplotlib documentation <matplotlib:api/animation_api>` for more details.

    Parameters
    ----------
    mad
        MADCompetition object whose synthesis we want to animate.
    framerate
        How many frames a second to display.
    batch_idx
        Which index to take from the batch dimension.
    channel_idx
        Which index to take from the channel dimension. If ``None``, we use all
        channels (assumed use-case is RGB(A) image).
    zoom
        How much to zoom in / enlarge the synthesized image, the ratio
        of display pixels to image pixels. If ``None``, we
        attempt to find the best value ourselves.
    fig
        If ``None``, we create a new figure. otherwise we assume this is
        an empty figure that has the appropriate size and number of
        subplots.
    axes_idx
        Dictionary specifying which axes contains which type of plot, allows
        for more fine-grained control of the resulting figure. Probably only
        helpful if fig is also defined. Possible keys: ``'mad_image',
        'loss', 'pixel_values', 'misc'``. Values should all be ints. If you
        tell this function to create a plot that doesn't have a corresponding
        key, we find the lowest int that is not already in the dict, so if you
        have axes that you want unchanged, place their idx in ``'misc'``.
    figsize
        The size of the figure to create. It may take a little bit of
        playing around to find a reasonable value. If ``None``, we attempt to
        make our best guess, aiming to have each axis be of size ``(5, 5)``.
    included_plots
        Which plots to include. Must be some subset of ``'display_mad_image',
        'plot_loss', 'plot_pixel_values'``.
    width_ratios
        If ``width_ratios`` is an empty dictionary, ``plot_loss`` will have
        double the width of the other plots. To change that, specify their
        relative widths using the keys: ['display_mad_image', 'plot_loss',
        'plot_pixel_values'] and floats specifying their relative width.

    Returns
    -------
    anim
        The animation object. In order to view, must convert to HTML
        or save.

    Raises
    ------
    ValueError
        If the given ``mad`` object was run with ``store_progress=False``.
    ValueError
        If ``mad.mage_image`` object is not 3d or 4d.
    ValueError
        If we do not know how to interpret the value of ``ylim``.

    Notes
    -----
    Unless specified, we use the ffmpeg backend, which requires that you have
    ffmpeg installed and on your path (https://ffmpeg.org/download.html).
    To use a different, use the matplotlib rcParams:
    ``matplotlib.rcParams['animation.writer'] = writer``, see
    `matplotlib documentation
    <https://matplotlib.org/stable/api/animation_api.html#writer-classes>`_ for more
    details.
    """
    if not mad.store_progress:
        raise ValueError(
            "synthesize() was run with store_progress=False, cannot animate!"
        )
    if mad.mad_image.ndim not in [3, 4]:
        raise ValueError(
            "animate() expects 3 or 4d data; unexpected behavior will result otherwise!"
        )
    _check_included_plots(included_plots, "included_plots")
    _check_included_plots(width_ratios, "width_ratios")
    _check_included_plots(axes_idx, "axes_idx")
    # we run plot_synthesis_status to initialize the figure if either fig is
    # None or if there are no titles on any axes, which we assume means that
    # it's an empty figure
    if fig is None or not any([ax.get_title() for ax in fig.axes]):
        fig, axes_idx = plot_synthesis_status(
            mad=mad,
            batch_idx=batch_idx,
            channel_idx=channel_idx,
            iteration=0,
            figsize=figsize,
            zoom=zoom,
            fig=fig,
            included_plots=included_plots,
            axes_idx=axes_idx,
            width_ratios=width_ratios,
        )
    # grab the artist for the second plot (we don't need to do this for the
    # MAD image plot, because we use the update_plot function for that)
    if "plot_loss" in included_plots:
        scat = [fig.axes[i].collections[0] for i in axes_idx["plot_loss"]]
    if "display_mad_image" in included_plots:
        fig.axes[axes_idx["display_mad_image"]].set_title("MAD Image")

    def movie_plot(i: int) -> list[mpl.artist.Artist]:
        """
        Matplotlib function for animation.

        Update plots for frame ``i``.

        Parameters
        ----------
        i
            The frame to plot.

        Returns
        -------
        artists
            The updated matplotlib artists.
        """
        # this warning is not relevant for animate
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="loss iteration and iteration for"
            )
            artists = []
            if "display_mad_image" in included_plots:
                artists.extend(
                    display.update_plot(
                        fig.axes[axes_idx["display_mad_image"]],
                        data=mad.saved_mad_image[i],
                        batch_idx=batch_idx,
                    )
                )
            if "plot_pixel_values" in included_plots:
                # this is the dumbest way to do this, but it's simple --
                # clearing the axes can cause problems if the user has, for
                # example, changed the tick locator or formatter. not sure how
                # to handle this best right now
                fig.axes[axes_idx["plot_pixel_values"]].clear()
                plot_pixel_values(
                    mad,
                    batch_idx=batch_idx,
                    channel_idx=channel_idx,
                    iteration=min(i * mad.store_progress, len(mad.losses) - 1),
                    ax=fig.axes[axes_idx["plot_pixel_values"]],
                )
            if "plot_loss" in included_plots:
                # loss always contains values from every iteration, but everything
                # else will be subsampled.
                x_val = mad._convert_iteration(i, False)
                scat[0].set_offsets((x_val, mad.reference_metric_loss[x_val]))
                scat[1].set_offsets((x_val, mad.optimized_metric_loss[x_val]))
                artists.extend(scat)
            # as long as blitting is True, need to return a sequence of artists
            return artists

    # don't need an init_func, since we handle initialization ourselves
    anim = mpl.animation.FuncAnimation(
        fig,
        movie_plot,
        frames=len(mad.saved_mad_image),
        blit=True,
        interval=1000.0 / framerate,
        repeat=False,
    )
    plt.close(fig)
    return anim


def display_mad_image_all(
    mad_metric1_min: MADCompetition,
    mad_metric2_min: MADCompetition,
    mad_metric1_max: MADCompetition,
    mad_metric2_max: MADCompetition,
    metric1_name: str | None = None,
    metric2_name: str | None = None,
    zoom: int | float = 1,
    **kwargs: Any,
) -> mpl.figure.Figure:
    """
    Display all MAD Competition images.

    To generate a full set of MAD Competition images, you need four instances:
    one for minimizing and maximizing each metric. This helper function creates
    a figure to display the full set of images.

    In addition to the four MAD Competition images, this also plots the initial
    image from ``mad_metric1_min``, for comparison.

    Parameters
    ----------
    mad_metric1_min
        ``MADCompetition`` object that minimized the first metric.
    mad_metric2_min
        ``MADCompetition`` object that minimized the second metric.
    mad_metric1_max
        ``MADCompetition`` object that maximized the first metric.
    mad_metric2_max
        ``MADCompetition`` object that maximized the second metric.
    metric1_name
        Name of the first metric. If ``None``, we use the name of the
        ``optimized_metric`` function from ``mad_metric1_min``.
    metric2_name
        Name of the second metric. If ``None``, we use the name of the
        ``optimized_metric`` function from ``mad_metric2_min``.
    zoom
        Ratio of display pixels to image pixels. See
        :func:`~plenoptic.tools.display.imshow` for details.
    **kwargs
        Passed to :func:`~plenoptic.tools.display.imshow`.

    Returns
    -------
    fig
        Figure containing the images.

    Raises
    ------
    ValueError
        If the four ``MADCompetition`` instances do not have the same ``image``
        attribute.
    """
    # this is a bit of a hack right now, because they don't all have same
    # initial image
    if not torch.allclose(mad_metric1_min.image, mad_metric2_min.image):
        raise ValueError("All four instances of MADCompetition must have same image!")
    if not torch.allclose(mad_metric1_min.image, mad_metric1_max.image):
        raise ValueError("All four instances of MADCompetition must have same image!")
    if not torch.allclose(mad_metric1_min.image, mad_metric2_max.image):
        raise ValueError("All four instances of MADCompetition must have same image!")
    if metric1_name is None:
        metric1_name = mad_metric1_min.optimized_metric.__name__
    if metric2_name is None:
        metric2_name = mad_metric2_min.optimized_metric.__name__
    fig = pt_make_figure(3, 2, [zoom * i for i in mad_metric1_min.image.shape[-2:]])
    mads = [mad_metric1_min, mad_metric1_max, mad_metric2_min, mad_metric2_max]
    titles = [
        f"Minimize {metric1_name}",
        f"Maximize {metric1_name}",
        f"Minimize {metric2_name}",
        f"Maximize {metric2_name}",
    ]
    # we're only plotting one image here, so if the user wants multiple
    # channels, they must be RGB
    if kwargs.get("channel_idx") is None and mad_metric1_min.initial_image.shape[1] > 1:
        as_rgb = True
    else:
        as_rgb = False
    display.imshow(
        mad_metric1_min.image,
        ax=fig.axes[0],
        title="Reference image",
        zoom=zoom,
        as_rgb=as_rgb,
        **kwargs,
    )
    display.imshow(
        mad_metric1_min.initial_image,
        ax=fig.axes[1],
        title="Initial (noisy) image",
        zoom=zoom,
        as_rgb=as_rgb,
        **kwargs,
    )
    for ax, mad, title in zip(fig.axes[2:], mads, titles):
        display_mad_image(mad, zoom=zoom, ax=ax, title=title, **kwargs)
    return fig


def plot_loss_all(
    mad_metric1_min: MADCompetition,
    mad_metric2_min: MADCompetition,
    mad_metric1_max: MADCompetition,
    mad_metric2_max: MADCompetition,
    metric1_name: str | None = None,
    metric2_name: str | None = None,
    metric1_kwargs: dict = {"c": "C0"},
    metric2_kwargs: dict = {"c": "C1"},
    min_kwargs: dict = {"linestyle": "--"},
    max_kwargs: dict = {"linestyle": "-"},
    figsize: tuple[int, int] = (10, 5),
) -> mpl.figure.Figure:
    """
    Plot loss for full set of MAD Competiton instances.

    To generate a full set of MAD Competition images, you need four instances:
    one for minimizing and maximizing each metric. This helper function creates
    a two-axis figure to display the loss for this full set.

    Parameters
    ----------
    mad_metric1_min
        ``MADCompetition`` object that minimized the first metric.
    mad_metric2_min
        ``MADCompetition`` object that minimized the second metric.
    mad_metric1_max
        ``MADCompetition`` object that maximized the first metric.
    mad_metric2_max
        ``MADCompetition`` object that maximized the second metric.
    metric1_name
        Name of the first metric. If ``None``, we use the name of the
        ``optimized_metric`` function from ``mad_metric1_min``.
    metric2_name
        Name of the second metric. If ``None``, we use the name of the
        ``optimized_metric`` function from ``mad_metric2_min``.
    metric1_kwargs
        Dictionary of arguments to pass to :func:`matplotlib.pyplot.plot` to identify
        synthesis instance where the first metric was being optimized.
    metric2_kwargs
        Dictionary of arguments to pass to :func:`matplotlib.pyplot.plot` to identify
        synthesis instance where the second metric was being optimized.
    min_kwargs
        Dictionary of arguments to pass to :func:`matplotlib.pyplot.plot` to identify
        synthesis instance where ``optimized_metric`` was being minimized.
    max_kwargs
        Dictionary of arguments to pass to :func:`matplotlib.pyplot.plot` to identify
        synthesis instance where ``optimized_metric`` was being maximized.
    figsize
        Size of the figure we create.

    Returns
    -------
    fig
        Figure containing the plot.

    Raises
    ------
    ValueError
        If the four ``MADCompetition`` instances do not have the same ``image``
        attribute.
    """
    if not torch.allclose(mad_metric1_min.image, mad_metric2_min.image):
        raise ValueError("All four instances of MADCompetition must have same image!")
    if not torch.allclose(mad_metric1_min.image, mad_metric1_max.image):
        raise ValueError("All four instances of MADCompetition must have same image!")
    if not torch.allclose(mad_metric1_min.image, mad_metric2_max.image):
        raise ValueError("All four instances of MADCompetition must have same image!")
    if metric1_name is None:
        metric1_name = mad_metric1_min.optimized_metric.__name__
    if metric2_name is None:
        metric2_name = mad_metric2_min.optimized_metric.__name__
    fig, axes = plt.subplots(1, 2, figsize=figsize)
    plot_loss(
        mad_metric1_min,
        axes=axes,
        label=f"Minimize {metric1_name}",
        **metric1_kwargs,
        **min_kwargs,
    )
    plot_loss(
        mad_metric1_max,
        axes=axes,
        label=f"Maximize {metric1_name}",
        **metric1_kwargs,
        **max_kwargs,
    )
    # we pass the axes backwards here because the fixed and synthesis metrics are
    # the opposite as they are in the instances above.
    plot_loss(
        mad_metric2_min,
        axes=axes[::-1],
        label=f"Minimize {metric2_name}",
        **metric2_kwargs,
        **min_kwargs,
    )
    plot_loss(
        mad_metric2_max,
        axes=axes[::-1],
        label=f"Maximize {metric2_name}",
        **metric2_kwargs,
        **max_kwargs,
    )
    axes[0].set(ylabel="Loss", title=metric2_name)
    axes[1].set(ylabel="Loss", title=metric1_name)
    axes[1].legend(loc="center left", bbox_to_anchor=(1.1, 0.5))
    return fig



import os.path as op
from contextlib import nullcontext as does_not_raise

import matplotlib.pyplot as plt
import pytest
import torch
from torch import nn

import plenoptic.synthesize.autodiff as autodiff
from conftest import DEVICE, ColorModel, get_model
from plenoptic.metric import mse
from plenoptic.simulate import Gaussian, OnOff
from plenoptic.simulate import LinearNonlinear as LNL
from plenoptic.synthesize import (
    MADCompetition,
    Metamer,
)
from plenoptic.synthesize.eigendistortion import (
    Eigendistortion,
    display_eigendistortion,
    display_eigendistortion_all,
)
from plenoptic.tools import examine_saved_synthesis, remove_grad, set_seed

# to be used for default model instantiation
SMALL_DIM = 20
LARGE_DIM = 100


class TestEigendistortionSynthesis:
    @pytest.mark.parametrize("model", ["frontend.OnOff.nograd"], indirect=True)
    def test_method_assertion(self, einstein_img, model):
        einstein_img = einstein_img[..., :SMALL_DIM, :SMALL_DIM]
        ed = Eigendistortion(einstein_img, model)
        with pytest.raises(ValueError, match="method must be in "):
            ed.synthesize(method="asdfsdfasf")

    @pytest.mark.parametrize(
        "model", ["frontend.OnOff.nograd", "ColorModel"], indirect=True
    )
    @pytest.mark.filterwarnings(
        "ignore:Jacobian > 1e6 elements and may cause out-of-memory:UserWarning"
    )
    def test_method_exact(self, model, einstein_img, color_img):
        # in this case, we're working with grayscale images
        if model.__class__ == OnOff:
            n_chans = 1
            img = einstein_img
        else:
            img = color_img
            n_chans = 3
        img = img[..., :SMALL_DIM, :SMALL_DIM]

        ed = Eigendistortion(img, model)
        # invert matrix explicitly
        ed.synthesize(method="exact")

        assert len(ed.eigenvalues) == n_chans * SMALL_DIM**2
        assert len(ed.eigendistortions) == n_chans * SMALL_DIM**2
        assert len(ed.eigenindex) == n_chans * SMALL_DIM**2

        # test that each eigenvector returned is original img shape
        assert ed.eigendistortions.shape[-3:] == (
            n_chans,
            SMALL_DIM,
            SMALL_DIM,
        )

    @pytest.mark.parametrize(
        "model", ["frontend.OnOff.nograd", "ColorModel"], indirect=True
    )
    def test_method_power(self, model, einstein_img, color_img):
        if model.__class__ == OnOff:
            n_chans = 1
            img = einstein_img
        else:
            img = color_img
            n_chans = 3
        img = img[..., :LARGE_DIM, :LARGE_DIM]
        ed = Eigendistortion(img, model)
        ed.synthesize(method="power", max_iter=3)

        # test it should only return two eigenvectors and values
        assert len(ed.eigenvalues) == 2
        assert len(ed.eigendistortions) == 2
        assert len(ed.eigenindex) == 2

        assert ed.eigendistortions.shape[-3:] == (
            n_chans,
            LARGE_DIM,
            LARGE_DIM,
        )

    @pytest.mark.parametrize("model", ["frontend.OnOff.nograd"], indirect=True)
    def test_orthog_iter(self, model, einstein_img):
        n, k = 30, 10
        n_chans = 1  # TODO color
        einstein_img = einstein_img[..., :n, :n]
        ed = Eigendistortion(einstein_img, model)
        ed.synthesize(k=k, method="power", max_iter=10)

        assert ed.eigendistortions.shape == (k * 2, n_chans, n, n)
        assert ed.eigenindex.allclose(
            torch.cat((torch.arange(k), torch.arange(n**2 - k, n**2)))
        )
        assert len(ed.eigenvalues) == 2 * k

    @pytest.mark.parametrize("model", ["frontend.OnOff.nograd"], indirect=True)
    @pytest.mark.filterwarnings(
        "ignore:Randomized SVD complete!:UserWarning",
    )
    def test_method_randomized_svd(self, model, einstein_img):
        n, k = 30, 10
        n_chans = 1  # TODO color
        einstein_img = einstein_img[..., :n, :n]
        ed = Eigendistortion(einstein_img, model)
        ed.synthesize(k=k, method="randomized_svd")
        assert ed.eigendistortions.shape == (k, n_chans, n, n)
        assert ed.eigenindex.allclose(torch.arange(k))
        assert len(ed.eigenvalues) == k

    @pytest.mark.parametrize("model", ["frontend.OnOff.nograd"], indirect=True)
    @pytest.mark.filterwarnings(
        "ignore:Randomized SVD complete!:UserWarning",
    )
    def test_method_accuracy(self, model, einstein_img):
        # test pow and svd against ground-truth jacobian (exact) method
        einstein_img = einstein_img[..., 125 : 125 + 25, 125 : 125 + 25]
        e_jac = Eigendistortion(einstein_img, model)
        e_pow = Eigendistortion(einstein_img, model)
        e_svd = Eigendistortion(einstein_img, model)

        k_pow, k_svd = 1, 75
        e_jac.synthesize(method="exact")
        set_seed(0)
        e_pow.synthesize(k=k_pow, method="power", max_iter=2500)
        e_svd.synthesize(k=k_svd, method="randomized_svd")

        print(
            "synthesized first and last: ",
            e_pow.eigenvalues[0],
            e_pow.eigenvalues[-1],
        )
        print(
            "exact first and last: ",
            e_jac.eigenvalues[0],
            e_jac.eigenvalues[-1],
        )

        assert e_pow.eigenvalues[0].isclose(e_jac.eigenvalues[0], atol=1e-2)
        assert e_pow.eigenvalues[-1].isclose(e_jac.eigenvalues[-1], atol=1e-2)
        assert e_svd.eigenvalues[0].isclose(e_jac.eigenvalues[0], atol=1e-2)

    @pytest.mark.parametrize(
        "model", ["frontend.OnOff.nograd", "ColorModel"], indirect=True
    )
    @pytest.mark.parametrize("method", ["power", "randomized_svd"])
    @pytest.mark.parametrize("k", [2, 3])
    @pytest.mark.filterwarnings(
        "ignore:Randomized SVD complete!:UserWarning",
    )
    def test_display(self, model, einstein_img, color_img, method, k):
        img = einstein_img if model.__class__ == OnOff else color_img
        as_rgb = model.__class__ == ColorModel

        img = img[..., :SMALL_DIM, :SMALL_DIM]
        eigendist = Eigendistortion(img, model)
        eigendist.synthesize(k=k, method=method, max_iter=10)
        display_eigendistortion(eigendist, eigenindex=0, as_rgb=as_rgb)
        display_eigendistortion(eigendist, eigenindex=1, as_rgb=as_rgb)

        if method == "power":
            display_eigendistortion(eigendist, eigenindex=-1, as_rgb=as_rgb)
            display_eigendistortion(eigendist, eigenindex=-2, as_rgb=as_rgb)
        elif method == "randomized_svd":  # svd only has top k not bottom k eigendists
            with pytest.raises(ValueError, match="eigenindex must be the index"):
                display_eigendistortion(eigendist, eigenindex=-1)
        plt.close("all")

    @pytest.mark.parametrize(
        "model", ["frontend.OnOff.nograd", "ColorModel"], indirect=True
    )
    @pytest.mark.parametrize("alpha", [1, [1], [1, 10], [1, 10, 10]])
    @pytest.mark.parametrize("eigenindex", [0, [0, -1], [0, -1, 5]])
    @pytest.mark.filterwarnings(
        "ignore:Adding 0.5 to distortion:UserWarning",
    )
    def test_display_all(self, model, einstein_img, color_img, alpha, eigenindex):
        # in this case, we're working with grayscale images
        img = einstein_img if model.__class__ == OnOff else color_img
        as_rgb = model.__class__ == ColorModel

        img = img[..., :SMALL_DIM, :SMALL_DIM]
        eigendist = Eigendistortion(img, model)
        eigendist.synthesize(k=2, method="power", max_iter=10)
        expectation = does_not_raise()
        if isinstance(eigenindex, list):
            if 5 in eigenindex:
                expectation = pytest.raises(
                    ValueError, match="eigenindex must be the index"
                )
            # this will get raised first
            if isinstance(alpha, list) and len(alpha) != len(eigenindex):
                expectation = pytest.raises(ValueError, match="If alpha is a list")
        elif isinstance(alpha, list) and len(alpha) != 1:
            expectation = pytest.raises(ValueError, match="If alpha is a list")
        with expectation:
            display_eigendistortion_all(
                eigendist, eigenindex=eigenindex, alpha=alpha, as_rgb=as_rgb
            )
        plt.close("all")

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("fail", [False, "img", "model"])
    @pytest.mark.parametrize("method", ["exact", "power", "randomized_svd"])
    @pytest.mark.filterwarnings(
        "ignore:Randomized SVD complete!:UserWarning",
    )
    def test_save_load(self, einstein_img, model, fail, method, tmp_path):
        if method in ["exact", "randomized_svd"]:
            img = einstein_img[..., :SMALL_DIM, :SMALL_DIM]
        else:
            img = einstein_img
        ed = Eigendistortion(img, model)
        ed.synthesize(max_iter=4, method=method)
        ed.save(op.join(tmp_path, "test_eigendistortion_save_load.pt"))
        if fail:
            if fail == "img":
                img = torch.rand_like(img)
                expectation = pytest.raises(
                    ValueError,
                    match="Saved and initialized attribute image have different values",
                )
            elif fail == "model":
                model = Gaussian(30).to(DEVICE)
                remove_grad(model)
                model.eval()
                expectation = pytest.raises(
                    ValueError,
                    match=("Saved and initialized model output have different values"),
                )
            ed_copy = Eigendistortion(img, model)
            with expectation:
                ed_copy.load(
                    op.join(tmp_path, "test_eigendistortion_save_load.pt"),
                    map_location=DEVICE,
                )
        else:
            ed_copy = Eigendistortion(img, model)
            ed_copy.load(
                op.join(tmp_path, "test_eigendistortion_save_load.pt"),
                map_location=DEVICE,
            )
            for k in ["image", "_representation_flat"]:
                if not getattr(ed, k).allclose(getattr(ed_copy, k), rtol=1e-2):
                    raise ValueError(
                        f"Something went wrong with saving and loading! {k} not"
                        " the same"
                    )
            # check that can resume
            ed_copy.synthesize(max_iter=4, method=method)

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    def test_load_init_fail(self, einstein_img, model, tmp_path):
        eig = Eigendistortion(einstein_img, model)
        eig.synthesize(max_iter=4)
        eig.save(op.join(tmp_path, "test_eigendistortion_load_init_fail.pt"))
        with pytest.raises(
            ValueError, match="load can only be called with a just-initialized"
        ):
            eig.load(op.join(tmp_path, "test_eigendistortion_load_init_fail.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("fail", [False, "name", "behavior"])
    def test_load_names(self, fail, einstein_img, model, tmp_path):
        # name and behavior same as our LinearNonlinear, but module path is
        # different
        if fail is False:

            class LinearNonlinear(torch.nn.Module):
                def __init__(self, *args, **kwargs):
                    super().__init__()
                    self.model = LNL((31, 31)).to(DEVICE)

                def forward(self, *args, **kwargs):
                    return self.model(*args, **kwargs)

            model2 = LinearNonlinear()
            expectation = does_not_raise()
        # name different but behavior same
        elif fail == "name":

            class LinearNonlinearFAIL(torch.nn.Module):
                def __init__(self, *args, **kwargs):
                    super().__init__()
                    self.model = LNL((31, 31)).to(DEVICE)

                def forward(self, *args, **kwargs):
                    return self.model(*args, **kwargs)

            model2 = LinearNonlinearFAIL()
            expectation = pytest.raises(
                ValueError, match="Saved and initialized model have different names"
            )
        # name same but behavior different
        elif fail == "behavior":

            class LinearNonlinear(torch.nn.Module):
                def __init__(self, *args, **kwargs):
                    super().__init__()
                    self.model = LNL((16, 16)).to(DEVICE)

                def forward(self, *args, **kwargs):
                    return self.model(*args, **kwargs)

            model2 = LinearNonlinear()
            expectation = pytest.raises(
                ValueError,
                match="Saved and initialized model output have different values",
            )
        eig = Eigendistortion(einstein_img, model)
        eig.synthesize(max_iter=4)
        eig.save(op.join(tmp_path, f"test_eigendistortion_load_names_{fail}.pt"))
        remove_grad(model2)
        model2.eval()
        eig = Eigendistortion(einstein_img, model2)
        with expectation:
            eig.load(op.join(tmp_path, f"test_eigendistortion_load_names_{fail}.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    def test_examine_saved_object(self, einstein_img, model, tmp_path):
        eig = Eigendistortion(einstein_img, model)
        eig.synthesize(max_iter=4)
        eig.save(op.join(tmp_path, "test_eigendistortion_examine.pt"))
        examine_saved_synthesis(op.join(tmp_path, "test_eigendistortion_examine.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("synth_type", ["met", "mad"])
    def test_load_object_type(self, einstein_img, model, synth_type, tmp_path):
        eig = Eigendistortion(einstein_img, model)
        eig.synthesize(max_iter=4)
        eig.save(op.join(tmp_path, "test_eigendistortion_load_object_type.pt"))
        if synth_type == "met":
            eig = Metamer(einstein_img, model)
        elif synth_type == "mad":
            eig = MADCompetition(
                einstein_img, mse, mse, "min", metric_tradeoff_lambda=1
            )
        with pytest.raises(
            ValueError, match="Saved object was a.* but initialized object is"
        ):
            eig.load(op.join(tmp_path, "test_eigendistortion_load_object_type.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("model_behav", ["dtype", "shape", "name"])
    def test_load_model_change(self, einstein_img, model, model_behav, tmp_path):
        eig = Eigendistortion(einstein_img, model)
        eig.synthesize(max_iter=4)
        eig.save(op.join(tmp_path, "test_eigendistortion_load_model_change.pt"))
        if model_behav == "dtype":
            # this actually gets raised in the model validation step (during init), not
            # load.
            expectation = pytest.raises(TypeError, match="model changes precision")
        elif model_behav == "shape":
            expectation = pytest.raises(
                ValueError,
                match="Saved and initialized model output have different shape",
            )
        elif model_behav == "name":
            expectation = pytest.raises(
                ValueError, match="Saved and initialized model have different names"
            )

        class NewModel(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.model = model

            def forward(self, x):
                if model_behav == "dtype":
                    return self.model(x).to(torch.float64)
                elif model_behav == "shape":
                    return self.model(x).flatten(-2)
                elif model_behav == "name":
                    return self.model(x)

        model = NewModel()
        model.eval()
        with expectation:
            eig = Eigendistortion(einstein_img, model)
            eig.load(op.join(tmp_path, "test_eigendistortion_load_model_change.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("attribute", ["saved", "init"])
    def test_load_attributes(self, einstein_img, model, attribute, tmp_path):
        eig = Eigendistortion(einstein_img, model)
        eig.synthesize(max_iter=4)
        if attribute == "saved":
            eig.test = "BAD"
            err_str = "Saved"
        eig.save(op.join(tmp_path, "test_eigendistortion_load_attributes.pt"))
        eig = Eigendistortion(einstein_img, model)
        if attribute == "init":
            eig.test = "BAD"
            err_str = "Initialized"
        with pytest.raises(
            ValueError, match=rf"{err_str} object has 1 attribute\(s\) not present"
        ):
            eig.load(op.join(tmp_path, "test_eigendistortion_load_attributes.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    def test_load_tol(self, einstein_img, model, tmp_path):
        eig = Eigendistortion(einstein_img, model)
        eig.synthesize(max_iter=5)
        eig.save(op.join(tmp_path, "test_eigendistortion_load_tol.pt"))
        eig = Eigendistortion(
            einstein_img + 1e-7 * torch.rand_like(einstein_img), model
        )
        with pytest.raises(ValueError, match="Saved and initialized attribute image"):
            eig.load(op.join(tmp_path, "test_eigendistortion_load_tol.pt"))
        eig.load(
            op.join(tmp_path, "test_eigendistortion_load_tol.pt"),
            tensor_equality_atol=1e-7,
        )

    @pytest.mark.parametrize(
        "model", ["naive.Identity", "NonModule", "frontend.OnOff.nograd"], indirect=True
    )
    @pytest.mark.parametrize("to_type", ["dtype", "device"])
    @pytest.mark.filterwarnings("ignore:Unable to call model.to:UserWarning")
    def test_to(self, curie_img, model, to_type):
        ed = Eigendistortion(curie_img, model)
        ed.synthesize(max_iter=5, method="power")
        if to_type == "dtype":
            # can't use the power method on a float16 tensor, so we use float64 instead
            # here.
            ed.to(torch.float64)
            assert ed.image.dtype == torch.float64
            assert ed.eigendistortions.dtype == torch.float64
        # can only run this one if we're on a device with CPU and GPU.
        elif to_type == "device" and DEVICE.type != "cpu":
            ed.to("cpu")
        ed.eigendistortions - ed.image
        ed.synthesize(max_iter=5, method="power")
        # reset so we don't mess up further tests
        if to_type == "dtype":
            ed.to(torch.float32)
        elif to_type == "device" and DEVICE.type != "cpu":
            ed.to(DEVICE)

    # test that we support models with 3d and 4d outputs
    @pytest.mark.parametrize(
        "model",
        ["PortillaSimoncelli", "frontend.LinearNonlinear.nograd"],
        indirect=True,
    )
    def test_model_dimensionality(self, einstein_img, model):
        eig = Eigendistortion(einstein_img, model)
        eig.synthesize(max_iter=5, method="power")

    @pytest.mark.parametrize(
        "model",
        [f"diff_dims-{i}" for i in range(1, 6)],
        indirect=True,
    )
    @pytest.mark.parametrize("input_dim", [3, 4, 5])
    @pytest.mark.filterwarnings("ignore:.*mostly been tested on 4d inputs:UserWarning")
    @pytest.mark.filterwarnings(
        "ignore:.*mostly been tested on models which:UserWarning"
    )
    def test_dimensionality(self, einstein_img, input_dim, model):
        img = einstein_img.squeeze()[..., :SMALL_DIM, :SMALL_DIM]
        while img.ndimension() < input_dim:
            img = img.unsqueeze(0)
        met = Eigendistortion(img, model)
        met.synthesize(max_iter=5, method="power")

    @pytest.mark.skipif(DEVICE.type == "cpu", reason="Only makes sense to test on cuda")
    @pytest.mark.parametrize("model", ["naive.Identity"], indirect=True)
    def test_map_location(self, curie_img, model, tmp_path):
        ed = Eigendistortion(curie_img, model)
        ed.synthesize(max_iter=4, method="power")
        ed.save(op.join(tmp_path, "test_eig_map_location.pt"))
        # calling load with map_location effectively switches everything
        # over to that device
        model.to("cpu")
        ed_copy = Eigendistortion(curie_img.to("cpu"), model)
        ed_copy.load(op.join(tmp_path, "test_eig_map_location.pt"), map_location="cpu")
        assert ed_copy.eigendistortions.device.type == "cpu"
        assert ed_copy.image.device.type == "cpu"
        ed_copy.synthesize(max_iter=4, method="power")
        # reset model device for other tests
        model.to(DEVICE)

    @pytest.mark.skipif(DEVICE.type == "cpu", reason="Only makes sense to test on cuda")
    @pytest.mark.parametrize("model", ["naive.Identity"], indirect=True)
    def test_to_midsynth(self, curie_img, model):
        ed = Eigendistortion(curie_img, model)
        ed.synthesize(max_iter=4, method="power")
        assert ed.eigendistortions.device.type == "cuda"
        assert ed.image.device.type == "cuda"
        ed.to("cpu")
        ed.synthesize(max_iter=4, method="power")
        assert ed.eigendistortions.device.type == "cpu"
        assert ed.image.device.type == "cpu"
        ed.to("cuda")
        ed.synthesize(max_iter=4, method="power")
        assert ed.eigendistortions.device.type == "cuda"
        assert ed.image.device.type == "cuda"

    @pytest.mark.parametrize("model", ["naive.Identity"], indirect=True)
    def test_change_precision_save_load(self, einstein_img, model, tmp_path):
        # Identity model doesn't change when you call .to() with a dtype
        # (unlike those models that have weights) so we use it here
        ed = Eigendistortion(einstein_img, model)
        ed.synthesize(max_iter=5)
        ed.to(torch.float64)
        assert ed.image.dtype == torch.float64, "dtype incorrect!"
        ed.save(op.join(tmp_path, "test_change_prec_save_load.pt"))
        ed_copy = Eigendistortion(einstein_img.to(torch.float64), model)
        ed_copy.load(op.join(tmp_path, "test_change_prec_save_load.pt"))
        ed_copy.synthesize(max_iter=5)
        assert ed_copy.image.dtype == torch.float64, "dtype incorrect!"


class TestAutodiffFunctions:
    @pytest.fixture(scope="class")
    def state(self, einstein_img):
        """variables to be reused across tests in this class"""

        k = 2  # num vectors with which to compute vjp, jvp, Fv
        einstein_img = einstein_img[
            ..., 100 : 100 + 16, 100 : 100 + 16
        ]  # reduce image size

        # eigendistortion object
        ed = Eigendistortion(einstein_img, get_model("frontend.OnOff.nograd"))

        x, y = ed._image_flat, ed._representation_flat

        x_dim = x.flatten().shape[0]
        y_dim = y.flatten().shape[0]

        return x, y, x_dim, y_dim, k

    def test_jacobian(self, state):
        x, y, x_dim, y_dim, k = state

        jac = autodiff.jacobian(y, x)
        assert jac.shape == (y_dim, x_dim)
        assert jac.requires_grad is False

    @pytest.mark.parametrize("detach", [False, True])
    def test_vec_jac_prod(self, state, detach):
        x, y, x_dim, y_dim, k = state

        U = torch.randn((y_dim, k), device=DEVICE)
        U = U / torch.linalg.vector_norm(U, ord=2, dim=0)

        vjp = autodiff.vector_jacobian_product(y, x, U, detach=detach)
        assert vjp.shape == (x_dim, k)
        assert vjp.requires_grad != detach

    def test_jac_vec_prod(self, state):
        x, y, x_dim, y_dim, k = state

        V = torch.randn((x_dim, k), device=DEVICE)
        V = V / torch.linalg.vector_norm(V, ord=2, dim=0)
        jvp = autodiff.jacobian_vector_product(y, x, V)
        assert jvp.shape == (y_dim, k)
        assert x.requires_grad and y.requires_grad
        assert jvp.requires_grad is False

    def test_fisher_vec_prod(self, state):
        x, y, x_dim, y_dim, k = state

        V, _ = torch.linalg.qr(torch.ones((x_dim, k), device=DEVICE), "reduced")
        U = V.clone()
        Jv = autodiff.jacobian_vector_product(y, x, V)
        Fv = autodiff.vector_jacobian_product(y, x, Jv)

        jac = autodiff.jacobian(y, x)

        Fv2 = jac.T @ jac @ U  # manually compute product to compare accuracy

        assert Fv.shape == (x_dim, k)
        assert Fv2.allclose(Fv, atol=1e-6)

    def test_simple_model_eigenvalues(self):
        """Test if Jacobian is constant in all directions for linear model"""
        singular_value = torch.ones(1, device=DEVICE) * 3.0

        class LM(nn.Module):
            """Simple y = Mx where M=3"""

            def __init__(self):
                super().__init__()
                self.M = nn.Linear(1, 1, bias=False)
                self.M.weight.data = singular_value

            def forward(self, x):
                y = self.M(x)
                return y

        x0 = torch.randn((1, 1, 5, 1), requires_grad=True, device=DEVICE)
        x0 = x0 / torch.linalg.vector_norm(x0, ord=2)
        mdl = LM().to(DEVICE)
        remove_grad(mdl)
        mdl.eval()

        k = 10
        x_dim = x0.numel()
        V = torch.randn((x_dim, k), device=DEVICE)  # random directions
        V = V / torch.linalg.vector_norm(V, ord=2, dim=0)

        e = Eigendistortion(x0, mdl)
        x, y = e._image_flat, e._representation_flat
        Jv = autodiff.jacobian_vector_product(y, x, V)
        Fv = autodiff.vector_jacobian_product(y, x, Jv)
        assert torch.diag(V.T @ Fv).sqrt().allclose(singular_value, rtol=1e-3)



import os

import einops
import numpy as np
import pytest
import torch

import plenoptic as po
from plenoptic.data.fetch import fetch_data

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
# if we have a second gpu, can use it for some tests
DEVICE2 = torch.device(1) if torch.cuda.device_count() > 1 else DEVICE
IMG_DIR = fetch_data("test_images.tar.gz")

# we do this to enable deterministic behavior on the gpu, for
# PortillaSimoncelli, see
# https://docs.nvidia.com/cuda/cublas/index.html#cublasApi_reproducibility for
# details
os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"

torch.set_num_threads(1)  # torch uses all avail threads which will slow tests
torch.manual_seed(0)
if torch.cuda.is_available():
    torch.cuda.manual_seed(0)


class ColorModel(torch.nn.Module):
    """Simple model that takes color image as input and outputs 2d conv."""

    def __init__(self):
        super().__init__()
        self.conv = torch.nn.Conv2d(3, 4, 3, 1)

    def forward(self, x):
        return self.conv(x)


@pytest.fixture(scope="package")
def curie_img():
    return po.load_images(IMG_DIR / "256" / "curie.pgm").to(DEVICE)


@pytest.fixture(scope="package")
def einstein_img():
    return po.load_images(IMG_DIR / "256" / "einstein.pgm").to(DEVICE)


@pytest.fixture(scope="package")
def einstein_img_small(einstein_img):
    return po.tools.center_crop(einstein_img, 64).to(DEVICE)


@pytest.fixture(scope="package")
def color_img():
    img = po.load_images(IMG_DIR / "256" / "color_wheel.jpg", as_gray=False).to(DEVICE)
    return img[..., :256, :256]


@pytest.fixture(scope="package")
def parrot_square():
    img = po.load_images(IMG_DIR / "mixed" / "Parrot.png").to(DEVICE)
    return po.tools.center_crop(img, 254)


@pytest.fixture(scope="package")
def parrot_square_double(parrot_square):
    return parrot_square.to(torch.float64)


@pytest.fixture(scope="package")
def einstein_img_double(einstein_img):
    return einstein_img.to(torch.float64)


@pytest.fixture(scope="package")
def basic_stim():
    return po.load_images(IMG_DIR / "256").to(DEVICE)


def get_model(name):
    if name == "SPyr":
        # in order to get a tensor back, need to wrap steerable pyramid so that
        # we can call convert_pyr_to_tensor in the forward call. in order for
        # that to work, downsample must be False
        class spyr(po.simul.SteerablePyramidFreq):
            def __init__(self, *args, **kwargs):
                kwargs.pop("downsample", None)
                super().__init__(*args, downsample=False, **kwargs)

            def forward(self, *args, **kwargs):
                coeffs = super().forward(*args, **kwargs)
                pyr_tensor, _ = po.simul.SteerablePyramidFreq.convert_pyr_to_tensor(
                    coeffs
                )
                return pyr_tensor

        # setting height=1 and # order=1 limits the size
        return spyr((256, 256), height=1, order=1).to(DEVICE)
    elif name == "LPyr":
        # in order to get a tensor back, need to wrap laplacian pyramid so that
        # we can flatten the output. in practice, not the best way to use this
        class lpyr(po.simul.LaplacianPyramid):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def forward(self, *args, **kwargs):
                coeffs = super().forward(*args, **kwargs)
                return torch.cat([c.flatten(-2) for c in coeffs], -1)

        return lpyr().to(DEVICE)
    elif name == "nlpd":
        return po.metric.nlpd
    elif name == "mse":
        return po.metric.naive.mse
    elif name == "ColorModel":
        model = ColorModel().to(DEVICE)
        po.tools.remove_grad(model)
        model.eval()
        return model

    # naive models
    elif name == "naive.Identity":
        model = po.simul.Identity().to(DEVICE)
        model.eval()
        return model
    elif name == "naive.CenterSurround":
        model = po.simul.CenterSurround((31, 31)).to(DEVICE)
        model.eval()
        return model
    elif name == "naive.Gaussian":
        model = po.simul.Gaussian((31, 31)).to(DEVICE)
        model.eval()
        return model
    elif name == "naive.Linear":
        model = po.simul.Linear((31, 31)).to(DEVICE)
        model.eval()
        return model

    # FrontEnd models:
    elif name == "frontend.LinearNonlinear":
        model = po.simul.LinearNonlinear((31, 31)).to(DEVICE)
        model.eval()
        return model
    elif name == "frontend.LinearNonlinear.nograd":
        model = po.simul.LinearNonlinear((31, 31)).to(DEVICE)
        po.tools.remove_grad(model)
        model.eval()
        return model
    elif name == "frontend.LuminanceGainControl":
        model = po.simul.LuminanceGainControl((31, 31)).to(DEVICE)
        model.eval()
        return model
    elif name == "frontend.LuminanceContrastGainControl":
        model = po.simul.LuminanceContrastGainControl((31, 31)).to(DEVICE)
        model.eval()
        return model
    elif name == "frontend.OnOff":
        model = po.simul.OnOff((31, 31), pretrained=True, cache_filt=True).to(DEVICE)
        model.eval()
        return model
    elif name == "frontend.OnOff.nograd":
        model = po.simul.OnOff((31, 31), pretrained=True, cache_filt=True).to(DEVICE)
        po.tools.remove_grad(model)
        model.eval()
        return model
    elif name == "frontend.OnOff.nograd.ctf":

        class OnOffCTF(po.simul.OnOff):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.scales = [0, 1]

            def forward(self, *args, scales=[], **kwargs):
                rep = super().forward(*args, **kwargs)
                if scales:
                    rep = rep[:, scales]
                return rep

        model = OnOffCTF((31, 31)).to(DEVICE)
        po.tools.remove_grad(model)
        model.eval()
        return model
    elif name == "VideoModel":
        # super simple model that combines across the batch dimension, as a
        # model with a temporal component would do
        class VideoModel(po.simul.OnOff):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def forward(self, *args, **kwargs):
                # this will do on/off on each batch separately
                rep = super().forward(*args, **kwargs)
                return rep.mean(0)

        model = VideoModel((31, 31), pretrained=True, cache_filt=True).to(DEVICE)
        po.tools.remove_grad(model)
        model.eval()
        return model
    elif name == "PortillaSimoncelli":
        return po.simul.PortillaSimoncelli((256, 256)).to(DEVICE)
    elif name == "NonModule":

        class NonModule:
            def __init__(self):
                self.name = "nonmodule"

            def __call__(self, x):
                return 1 * x

        return NonModule()
    elif "diff_dims" in name:

        class DimModel(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.target_dims = int(name.replace("diff_dims-", ""))

            def forward(self, img):
                img = img.flatten()
                while img.ndimension() != self.target_dims:
                    img = img.unsqueeze(0)
                return img

        model = DimModel()
        model.eval()
        return model


@pytest.fixture(scope="package")
def model(request):
    return get_model(request.param)


# this is the same as model() fixture above, in order to get two independent
# fixtures.
@pytest.fixture(scope="package")
def model2(request):
    return get_model(request.param)


def check_loss_saved_synth(
    losses, saved_synth, target_iter, objective_function, store_progress
):
    assert len(saved_synth) == np.ceil(target_iter / store_progress) + 1, (
        "Didn't end up with enough saved synth after first synth!"
    )
    assert len(losses) == target_iter + 1, (
        "Didn't end up with enough losses after first synth!"
    )
    if store_progress is True:
        losses = losses
    elif (target_iter % store_progress) == 0:
        losses = losses[::store_progress]
    else:
        # then we need to add on the loss from the current synth object.
        losses = einops.pack([losses[::store_progress], losses[-1]], "*")[0]
    assert len(losses) == len(saved_synth), "wrong length!"
    for synth_loss, saved in zip(losses.to(DEVICE), saved_synth):
        loss = objective_function(saved.to(DEVICE)).squeeze()
        if not torch.equal(loss, synth_loss):
            raise ValueError("saved_synth and loss are misaligned!")



"""
Tools to deal with data from outside plenoptic.

For example, images synthesized using the code from another paper.
"""

import os.path as op
from typing import Any

import imageio.v3 as iio
import matplotlib as mpl
import matplotlib.lines as lines
import numpy as np
import pyrtools as pt
import scipy.io as sio

from ..data.fetch import fetch_data


def plot_MAD_results(
    original_image: str,
    noise_levels: list[int] | None = None,
    results_dir: str | None = None,
    ssim_images_dir: str | None = None,
    zoom: int | float = 3,
    vrange: str = "indep1",
    **kwargs: Any,
) -> tuple[mpl.figure.Figure, dict[str, dict[str, float | np.ndarray]]]:
    r"""
    Plot original MAD results, provided by Zhou Wang.

    Plot the results of original MAD Competition, as provided in .mat
    files. The figure created shows the results for one reference image
    and multiple noise levels. The reference image is plotted on the
    first row, followed by a separate row for each noise level, which
    will show the initial (noisy) image and the four synthesized images,
    with their respective losses for the two metrics (MSE and SSIM).

    We also return a dictionary that contains the losses, noise levels, and original
    image name for each plotted noise level.

    This code can probably be adapted to other uses, but requires that
    all images are the same size and assumes they're all 64 x 64 pixels.

    Parameters
    ----------
    original_image
        Which of the sample images to plot. Must be of the form ``f"samp{i}"``
        where ``i`` is an integer between 1 and 10 (inclusive).
    noise_levels
        Which noise levels to plot. if ``None``, will plot all. If a list,
        elements must be :math:`2^i` where :math:`i\in [1, 10]`.
    results_dir
        Path to the results directory containing the results.mat files. If
        ``None``, we download them (requires optional dependency ``pooch``).
    ssim_images_dir
        Path to the directory containing the .tif images used in SSIM paper. If
        ``None``, we download them (requires optional dependency ``pooch``).
    zoom
        Ratio of display pixels to image pixels, passed to
        :func:`~plenoptic.tools.display.imshow`. If >1, must be an integer. If <1, must
        be 1/d where d is a a divisor of the size of the largest image.
    vrange
        How to map image values to colormap. In addition to the values accepted by
        :func:`~plenoptic.tools.display.imshow`, we also accept ``"row0/1/2/3"``, which
        is the same as ``"auto0/1/2/3"``, except that we do it on a per-row basis (all
        images with same noise level).
    **kwargs
        Passed to :func:`~plenoptic.tools.display.imshow`. Note that we call imshow
        separately on each image and so any argument that relies on imshow having
        access to all images will probably not work as expected.

    Returns
    -------
    fig :
        Figure containing the images.
    results :
        Dictionary containing the errors for each noise level. To
        convert to a well-structured pandas DataFrame, run
        ``pandas.DataFrame(results).T``.

    Raises
    ------
    ValueError
        If ``original_image`` takes an illegal value.
    """
    if results_dir is None:
        results_dir = str(fetch_data("MAD_results.tar.gz"))
    if ssim_images_dir is None:
        ssim_images_dir = str(fetch_data("ssim_images.tar.gz"))
    allowed_vals = [f"samp{i}" for i in range(1, 11)]
    if original_image not in allowed_vals:
        err_msg = f"original_image must be one of {allowed_vals}"
        raise ValueError(err_msg)
    img_path = op.join(op.expanduser(ssim_images_dir), f"{original_image}.tif")
    orig_img = iio.imread(img_path)
    blanks = np.ones((*orig_img.shape, 4))
    if noise_levels is None:
        noise_levels = [2**i for i in range(1, 11)]
    results = {}
    images = np.dstack([orig_img, blanks])
    titles = ["Original image"] + 4 * [None]
    super_titles = 5 * [None]
    keys = [
        "im_init",
        "im_fixmse_maxssim",
        "im_fixmse_minssim",
        "im_fixssim_minmse",
        "im_fixssim_maxmse",
    ]
    for level in noise_levels:
        mat = sio.loadmat(
            op.join(
                op.expanduser(results_dir),
                f"{original_image}_L{level}_results.mat",
            ),
            squeeze_me=True,
        )
        # remove these metadata keys
        [mat.pop(k) for k in ["__header__", "__version__", "__globals__"]]
        key_titles = [
            f"Noise level: {level}",
            f"Best SSIM: {mat['maxssim']:.05f}",
            f"Worst SSIM: {mat['minssim']:.05f}",
            f"Best MSE: {mat['minmse']:.05f}",
            f"Worst MSE: {mat['maxmse']:.05f}",
        ]
        key_super_titles = [
            None,
            f"Fix MSE: {mat['FIX_MSE']:.0f}",
            None,
            f"Fix SSIM: {mat['FIX_SSIM']:.05f}",
            None,
        ]
        for k, t, s in zip(keys, key_titles, key_super_titles):
            images = np.dstack([images, mat.pop(k)])
            titles.append(t)
            super_titles.append(s)
        # this then just contains the loss information
        mat.update({"noise_level": level, "original_image": original_image})
        results[f"L{level}"] = mat
    images = images.transpose((2, 0, 1))
    if vrange.startswith("row"):
        vrange_list = []
        for i in range(len(images) // 5):
            vr, cmap = pt.tools.display.colormap_range(
                images[5 * i : 5 * (i + 1)], vrange.replace("row", "auto")
            )
            vrange_list.extend(vr)
    else:
        vrange_list, cmap = pt.tools.display.colormap_range(images, vrange)
    # this is a bit of hack to do the same thing imshow does, but with
    # slightly more space dedicated to the title
    fig = pt.tools.display.make_figure(
        len(images) // 5,
        5,
        [zoom * i + 1 for i in images.shape[-2:]],
        vert_pct=0.75,
    )
    for img, ax, t, vr, s in zip(images, fig.axes, titles, vrange_list, super_titles):
        # these are the blanks
        if (img == 1).all():
            continue
        pt.imshow(img, ax=ax, title=t, zoom=zoom, vrange=vr, cmap=cmap, **kwargs)
        if s is not None:
            font = {
                k.replace("_", ""): v
                for k, v in ax.title.get_font_properties().__dict__.items()
            }
            # these are the acceptable keys for the fontdict below
            font = {
                k: v
                for k, v in font.items()
                if k in ["family", "color", "weight", "size", "style"]
            }
            # for some reason, this (with passing the transform) is
            # different (and looks better) than using ax.text. We also
            # slightly adjust the placement of the text to account for
            # different zoom levels (we also have 10 pixels between the
            # rows and columns, which correspond to a different)
            img_size = ax.bbox.size
            fig.text(
                1 + (5 / img_size[0]),
                (1 / 0.75),
                s,
                fontdict=font,
                transform=ax.transAxes,
                ha="center",
                va="top",
            )
            # linewidth of 1.5 looks good with bbox of 192, 192
            linewidth = np.max([1.5 * np.mean(img_size / 192), 1])
            line = lines.Line2D(
                2 * [0 - ((5 + linewidth / 2) / img_size[0])],
                [0, (1 / 0.75)],
                transform=ax.transAxes,
                figure=fig,
                linewidth=linewidth,
            )
            fig.lines.append(line)
    return fig, results



"""
Test uploaded files.

For the documentation, we have pre-generated some synthesis outputs and uplodaed
them to OSF. Then during docs build, we download them.

These tests check that the outputs haven't changed, notifying us if we need to
update them.
"""

import os
from collections import OrderedDict

import einops
import numpy as np
import pytest
import torch
from torchvision import models
from torchvision.models import feature_extraction

import plenoptic as po
from conftest import DEVICE, DEVICE2
from plenoptic.data.fetch import fetch_data
from plenoptic.tools.data import _check_tensor_equality


def compare_eigendistortions(eig, eig_up, rtol=1e-5, atol=1e-7):
    for k in ["_representation_flat", "eigendistortions"]:
        _check_tensor_equality(
            getattr(eig, k),
            getattr(eig_up, k),
            "Test",
            "OSF",
            rtol,
            atol,
            f"{k} has different {{error_type}}! Update the OSF version.",
        )


def compare_metamers(met, met_up, rtol=1e-5, atol=1e-7):
    _check_tensor_equality(
        met.metamer,
        met_up.metamer,
        "Test",
        "OSF",
        rtol,
        atol,
        "metamer has different {error_type}! Update the OSF version.",
    )


class PortillaSimoncelliRemove(po.simul.PortillaSimoncelli):
    r"""Model for measuring a subset of texture statistics reported by
    PortillaSimoncelli

    Parameters
    ----------
    im_shape: int
        the size of the images being processed by the model
    remove_keys: list
        The dictionary keys for the statistics we will "remove".  In practice we set
        them to zero.
        Possible keys: ["pixel_statistics", "auto_correlation_magnitude",
        "skew_reconstructed", "kurtosis_reconstructed",
        "auto_correlation_reconstructed", "std_reconstructed", "magnitude_std",
        "cross_orientation_correlation_magnitude", "cross_scale_correlation_magnitude",
        "cross_scale_correlation_real", "var_highpass_residual"]
    """

    def __init__(
        self,
        im_shape,
        remove_keys,
    ):
        super().__init__(im_shape, n_scales=4, n_orientations=4, spatial_corr_width=9)
        self.remove_keys = remove_keys

    def forward(self, image, scales=None):
        r"""Generate Texture Statistics representation of an image with `remove_keys`
        removed.

        Parameters
        ----------
        image : torch.Tensor
            A tensor containing the image to analyze.
        scales : list, optional
            Which scales to include in the returned representation. If an empty
            list (the default), we include all scales. Otherwise, can contain
            subset of values present in this model's scales attribute.

        Returns
        -------
        representation: torch.Tensor
            3d tensor of shape (batch, channel, stats) containing the measured texture
            stats.

        """
        # create the representation tensor (with all scales)
        stats_vec = super().forward(image)
        # convert to dict so it's easy to zero out the keys we don't care about
        stats_dict = self.convert_to_dict(stats_vec)
        for kk in self.remove_keys:
            # we zero out the stats (instead of removing them) because removing them
            # makes it difficult to keep track of which stats belong to which scale
            # (which is necessary for coarse-to-fine synthesis) -- see discussion above.
            if isinstance(stats_dict[kk], OrderedDict):
                for key, val in stats_dict[kk].items():
                    stats_dict[kk][key] *= 0
            else:
                stats_dict[kk] *= 0
        # then convert back to tensor and remove any scales we don't want
        # (for coarse-to-fine)  -- see discussion above.
        stats_vec = self.convert_to_tensor(stats_dict)
        if scales is not None:
            stats_vec = self.remove_scales(stats_vec, scales)
        return stats_vec


class PortillaSimoncelliMask(po.simul.PortillaSimoncelli):
    r"""Extend the PortillaSimoncelli model to operate on masked images.

    Additional Parameters
    ----------
    mask: Tensor
        boolean mask with ``True`` in the part of the image that will be filled in
        during synthesis
    target: Tensor
        image target for synthesis

    """

    def __init__(
        self,
        im_shape,
        n_scales=4,
        n_orientations=4,
        spatial_corr_width=9,
        mask=None,
        target=None,
    ):
        super().__init__(im_shape, n_scales=4, n_orientations=4, spatial_corr_width=9)
        self.mask = mask
        self.target = target

    def forward(self, image, scales=None):
        r"""Generate Texture Statistics representation of an image using the target for
        the masked portion

        Parameters
        ----------
        images : torch.Tensor
            A 4d tensor containing two images to analyze, with shape (2,
            channel, height, width).
        scales : list, optional
            Which scales to include in the returned representation. If an empty
            list (the default), we include all scales. Otherwise, can contain
            subset of values present in this model's scales attribute.

        Returns
        -------
        representation_tensor: torch.Tensor
            3d tensor of shape (batch, channel, stats) containing the measured
            texture statistics.

        """
        if self.mask is not None and self.target is not None:
            image = self.texture_masked_image(image)

        return super().forward(image, scales=scales)

    def texture_masked_image(self, image):
        r"""Fill in part of the image (designated by the mask) with the saved target
        image

        Parameters
        ------------
        image : torch.Tensor
            A tensor containing a single image

        Returns
        -------
        texture_masked_image: torch.Tensor
            An image that is a combination of the input image and the saved target.
            Combination is specified by self.mask

        """
        return self.target * self.mask + image * (~self.mask)


class PortillaSimoncelliMixture(po.simul.PortillaSimoncelli):
    r"""Extend the PortillaSimoncelli model to mix two different images

    Parameters
    ----------
    im_shape: int
        the size of the images being processed by the model

    """

    def __init__(
        self,
        im_shape,
    ):
        super().__init__(im_shape, n_scales=4, n_orientations=4, spatial_corr_width=9)

    def forward(self, images, scales=None):
        r"""Average Texture Statistics representations of two image

        Parameters
        ----------
        images : torch.Tensor
            A 4d tensor containing one or two images to analyze, with shape (i,
            channel, height, width), i in {1,2}.
        scales : list, optional
            Which scales to include in the returned representation. If an empty
            list (the default), we include all scales. Otherwise, can contain
            subset of values present in this model's scales attribute.

        Returns
        -------
        representation_tensor: torch.Tensor
            3d tensor of shape (batch, channel, stats) containing the measured
            texture statistics.

        """
        if images.shape[0] == 2:
            # need the images to be 4d, so we use the "1 element slice"
            stats0 = super().forward(images[:1], scales=scales)
            stats1 = super().forward(images[1:2], scales=scales)
            return (stats0 + stats1) / 2
        else:
            return super().forward(images, scales=scales)


class PortillaSimoncelliMagMeans(po.simul.PortillaSimoncelli):
    r"""Include the magnitude means in the PS texture representation.

    Parameters
    ----------
    im_shape: int
        the size of the images being processed by the model

    """

    def __init__(
        self,
        im_shape,
    ):
        super().__init__(im_shape, n_scales=4, n_orientations=4, spatial_corr_width=7)

    def forward(self, image, scales=None):
        r"""Average Texture Statistics representations of two image

        Parameters
        ----------
        image : torch.Tensor
            A 4d tensor (batch, channel, height, width) containing the image(s) to
            analyze.
        scales : list, optional
            Which scales to include in the returned representation. If an empty
            list (the default), we include all scales. Otherwise, can contain
            subset of values present in this model's scales attribute.

        Returns
        -------
        representation_tensor: torch.Tensor
            3d tensor of shape (batch, channel, stats) containing the measured
            texture statistics.

        """
        stats = super().forward(image, scales=scales)
        # this helper function returns a list of tensors containing the steerable
        # pyramid coefficients at each scale
        pyr_coeffs = self._compute_pyr_coeffs(image)[1]
        # only compute the magnitudes for the desired scales
        magnitude_pyr_coeffs = [
            coeff.abs()
            for i, coeff in enumerate(pyr_coeffs)
            if scales is None or i in scales
        ]
        magnitude_means = [mag.mean((-2, -1)) for mag in magnitude_pyr_coeffs]
        return einops.pack([stats, *magnitude_means], "b c *")[0]

    # overwriting these following two methods allows us to use the plot_representation
    # method with the modified model, making examining it easier.
    def convert_to_dict(self, representation_tensor: torch.Tensor) -> OrderedDict:
        """Convert tensor of stats to dictionary."""
        n_mag_means = self.n_scales * self.n_orientations
        rep = super().convert_to_dict(representation_tensor[..., :-n_mag_means])
        mag_means = representation_tensor[..., -n_mag_means:]
        rep["magnitude_means"] = einops.rearrange(
            mag_means,
            "b c (s o) -> b c s o",
            s=self.n_scales,
            o=self.n_orientations,
        )
        return rep

    def _representation_for_plotting(self, rep: OrderedDict) -> OrderedDict:
        r"""Convert the data into a dictionary representation that is more convenient
        for plotting.

        Intended as a helper function for plot_representation.
        """
        mag_means = rep.pop("magnitude_means")
        data = super()._representation_for_plotting(rep)
        data["magnitude_means"] = mag_means.flatten()
        return data


@pytest.mark.order(1)
@pytest.mark.skipif(DEVICE.type == "cpu", reason="Only do this on cuda")
class TestDoctest:
    def test_eigendistortion(self, einstein_img_double):
        torch.use_deterministic_algorithms(True)
        po.tools.set_seed(0)
        os.makedirs("uploaded_files", exist_ok=True)
        torch.save(
            torch.random.get_rng_state(),
            "uploaded_files/torch_rng_state_eigendistortion.pt",
        )
        print(np.random.get_state())
        lg = po.simul.LuminanceGainControl(
            (31, 31), pad_mode="circular", pretrained=True, cache_filt=True
        )
        po.tools.remove_grad(lg)
        lg = lg.to(DEVICE).to(einstein_img_double.dtype)
        lg.eval()
        eig = po.synth.Eigendistortion(einstein_img_double, lg)
        eig.synthesize(max_iter=1000)
        eig.save("uploaded_files/example_eigendistortion.pt")
        eig_up = po.synth.Eigendistortion(einstein_img_double, lg)
        eig_up.load(
            fetch_data("example_eigendistortion.pt"),
            tensor_equality_atol=1e-7,
            map_location=DEVICE,
        )
        compare_eigendistortions(eig, eig_up)


@pytest.mark.order(0)
@pytest.mark.skipif(DEVICE.type == "cpu", reason="Only do this on cuda")
@pytest.mark.skipif(
    os.environ.get("RUN_REGRESSION_SYNTH", "") != "1",
    reason="These take a long time, so don't run every time",
)
class TestTutorialNotebooks:
    class TestDemoEigendistortion:
        def test_berardino_onoff(self, parrot_square_double):
            torch.use_deterministic_algorithms(True)
            po.tools.set_seed(0)
            os.makedirs("uploaded_files", exist_ok=True)
            torch.save(
                torch.random.get_rng_state(),
                "uploaded_files/torch_rng_state_berardino_onoff.pt",
            )
            print(np.random.get_state())
            model = po.simul.OnOff(
                (31, 31),
                pretrained=True,
                cache_filt=True,
                apply_mask=True,
            )
            po.tools.remove_grad(model)
            model = model.to(DEVICE).to(parrot_square_double.dtype)
            model.eval()
            eig = po.synth.Eigendistortion(parrot_square_double, model)
            eig.synthesize(k=3, method="power", max_iter=2000)
            eig.save("uploaded_files/berardino_onoff.pt")
            eig_up = po.synth.Eigendistortion(parrot_square_double, model)
            eig_up.load(
                fetch_data("berardino_onoff.pt"),
                tensor_equality_atol=1e-7,
                map_location=DEVICE,
            )
            compare_eigendistortions(eig, eig_up)

        def test_berardino_vgg16(self, parrot_square_double):
            # Create a class that takes the nth layer output of a given model
            class TorchVision(torch.nn.Module):
                def __init__(self, model, return_node: str):
                    super().__init__()
                    self.extractor = feature_extraction.create_feature_extractor(
                        model, return_nodes=[return_node]
                    )
                    self.model = model
                    self.return_node = return_node

                def forward(self, x):
                    return self.extractor(x)[self.return_node]

            def normalize(img_tensor):
                """standardize the image for vgg16"""
                return (img_tensor - img_tensor.mean()) / img_tensor.std()

            torch.use_deterministic_algorithms(True)
            po.tools.set_seed(0)
            os.makedirs("uploaded_files", exist_ok=True)
            torch.save(
                torch.random.get_rng_state(),
                "uploaded_files/torch_rng_state_berardino_vgg16.pt",
            )
            print(np.random.get_state())
            model = models.vgg16(
                weights=models.VGG16_Weights.IMAGENET1K_V1, progress=False
            )
            model = TorchVision(model, "features.11")
            po.tools.remove_grad(model)
            img = parrot_square_double.to(DEVICE2)
            img = normalize(img).repeat(1, 3, 1, 1)
            model = model.to(DEVICE2).to(img.dtype)
            model.eval()
            eig = po.synth.Eigendistortion(img, model)
            eig.synthesize(k=2, method="power", max_iter=5000)
            eig.save("uploaded_files/berardino_vgg16.pt")
            eig_up = po.synth.Eigendistortion(img, model)
            eig_up.load(
                fetch_data("berardino_vgg16.pt"),
                tensor_equality_atol=1e-7,
                map_location=DEVICE2,
            )
            compare_eigendistortions(eig, eig_up)

    @pytest.mark.filterwarnings(
        "ignore:Loss has converged, stopping synthesis:UserWarning"
    )
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    class TestPortillaSimoncelli:
        @pytest.fixture(scope="class")
        def ps_images(self):
            img_dir = fetch_data("portilla_simoncelli_images.tar.gz")
            images = po.load_images(img_dir).to(DEVICE).to(torch.float64)
            filenames = [p.stem for p in sorted(img_dir.iterdir())]
            return images, filenames

        def get_specific_img(self, images, filenames, tgt_filename):
            # the clone is here because torch saving/loading preserves views and
            # so extra info would be saved without it:
            # https://docs.pytorch.org/docs/stable/notes/serialization.html#saving-and-loading-tensors-preserves-views
            return images[filenames.index(tgt_filename)].unsqueeze(0).clone()

        @pytest.fixture(scope="class")
        def ps_regression(self):
            return fetch_data("ps_regression.tar.gz")

        @pytest.mark.parametrize(
            "fn",
            [
                "fig4a",
                "fig12a",
                "fig12b",
                "fig12c",
                "fig12d",
                "fig12e",
                "fig12f",
                "fig13a",
                "fig13b",
                "fig13c",
                "fig13d",
                "fig14a",
                "fig14b",
                "fig14c",
                "fig14d",
                "fig14e",
                "fig15a",
                "fig15b",
                "fig15c",
                "fig15d",
                "fig15e",
                "fig15f",
                "fig16a",
                "fig16b",
                "fig16c",
                "fig16d",
                "fig16e",
                "fig16f",
                "fig18a",
                "einstein",
            ],
        )
        def test_ps_basic_synthesis(
            self, ps_images, fn, einstein_img_double, ps_regression
        ):
            torch.use_deterministic_algorithms(True)
            po.tools.set_seed(0)
            torch.save(
                torch.random.get_rng_state(),
                f"uploaded_files/torch_rng_state_ps_basic_{fn}.pt",
            )
            print(np.random.get_state())
            if fn.startswith("fig"):
                img = self.get_specific_img(*ps_images, fn)
            elif fn == "einstein":
                img = einstein_img_double
            # this is a sawtooth grating, with 4 scales the steerable pyramid's
            # residual lowpass is uniform and thus correlation between it and
            # the coarsest scale is all NaNs (i.e., the last scale of
            # auto_correlation_reconstructed is all NaNs)
            n_scales = 3 if fn == "fig12b" else 4
            model = po.simul.PortillaSimoncelli(img.shape[-2:], n_scales=n_scales)
            model.to(DEVICE).to(torch.float64)
            met = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            met.setup(
                ((torch.rand_like(img) - 0.5) * 0.1 + img.mean()).clip(min=0, max=1)
            )
            met.synthesize(
                max_iter=3000,
                change_scale_criterion=None,
                ctf_iters_to_check=7,
            )
            met.save(f"uploaded_files/ps_basic_synthesis_{fn}.pt")
            met_up = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            met_up.load(
                ps_regression / f"ps_basic_synthesis_{fn}.pt",
                tensor_equality_atol=1e-7,
                map_location=DEVICE,
            )
            compare_metamers(met, met_up)

        # make sure we fail if save load with different stats removed
        def test_ps_remove_fail(self, ps_images, tmp_path):
            img = self.get_specific_img(*ps_images, "fig4a")
            model = PortillaSimoncelliRemove(
                img.shape[-2:], remove_keys=["pixel_statistics"]
            )
            model.to(DEVICE).to(torch.float64)
            im_init = (torch.rand_like(img) - 0.5) * 0.1 + img.mean()
            met = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            met.setup(im_init)
            met.synthesize(
                max_iter=5, change_scale_criterion=None, ctf_iters_to_check=7
            )
            met.save(tmp_path / "test_ps_remove_fail.pt")
            model = PortillaSimoncelliRemove(
                img.shape[-2:], remove_keys=["pixel_statistics", "skew_reconstructed"]
            )
            model.to(DEVICE).to(torch.float64)
            met = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            with pytest.raises(ValueError, match="Saved and initialized model output"):
                met.load(tmp_path / "test_ps_remove_fail.pt", map_location=DEVICE)

        @pytest.mark.parametrize(
            "fn, stats",
            [
                (
                    "fig3a",
                    [
                        "pixel_statistics",
                        "skew_reconstructed",
                        "kurtosis_reconstructed",
                    ],
                ),
                (
                    "fig3b",
                    [
                        "pixel_statistics",
                        "skew_reconstructed",
                        "kurtosis_reconstructed",
                    ],
                ),
                ("fig4a", ["auto_correlation_reconstructed", "std_reconstructed"]),
                ("fig4b", ["auto_correlation_reconstructed", "std_reconstructed"]),
                (
                    "fig6a",
                    [
                        "magnitude_std",
                        "cross_orientation_correlation_magnitude",
                        "cross_scale_correlation_magnitude",
                        "auto_correlation_magnitude",
                    ],
                ),
                (
                    "fig6b",
                    [
                        "magnitude_std",
                        "cross_orientation_correlation_magnitude",
                        "cross_scale_correlation_magnitude",
                        "auto_correlation_magnitude",
                    ],
                ),
                ("fig8a", ["cross_scale_correlation_real"]),
                ("fig8b", ["cross_scale_correlation_real"]),
            ],
        )
        @pytest.mark.parametrize("remove_bool", [True, False])
        def test_ps_remove(self, ps_images, fn, stats, remove_bool, ps_regression):
            torch.use_deterministic_algorithms(True)
            po.tools.set_seed(0)
            torch.save(
                torch.random.get_rng_state(),
                f"uploaded_files/torch_rng_state_ps_remove_{fn}_remove-{remove_bool}.pt",
            )
            print(np.random.get_state())
            img = self.get_specific_img(*ps_images, fn).to(DEVICE2)
            if remove_bool:
                model = PortillaSimoncelliRemove(img.shape[-2:], remove_keys=stats)
            else:
                model = po.simul.PortillaSimoncelli(img.shape[-2:])
            model.to(DEVICE2).to(torch.float64)
            met = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            met.setup((torch.rand_like(img) - 0.5) * 0.1 + img.mean())
            met.synthesize(
                max_iter=3000, change_scale_criterion=None, ctf_iters_to_check=7
            )
            met.save(f"uploaded_files/ps_remove_{fn}_remove-{remove_bool}.pt")
            met_up = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            met_up.load(
                ps_regression / f"ps_remove_{fn}_remove-{remove_bool}.pt",
                tensor_equality_atol=1e-7,
                map_location=DEVICE2,
            )
            compare_metamers(met, met_up)

        @pytest.mark.filterwarnings("ignore:You will need to call setup:UserWarning")
        def test_ps_mask(self, ps_images, ps_regression):
            torch.use_deterministic_algorithms(True)
            po.tools.set_seed(0)
            torch.save(
                torch.random.get_rng_state(),
                "uploaded_files/torch_rng_state_ps_mask.pt",
            )
            print(np.random.get_state())
            img = self.get_specific_img(*ps_images, "fig14b").to(DEVICE2)
            mask = torch.zeros_like(img).bool()
            ctr_dim = (img.shape[-2] // 4, img.shape[-1] // 4)
            mask[..., ctr_dim[0] : 3 * ctr_dim[0], ctr_dim[1] : 3 * ctr_dim[1]] = True
            model = PortillaSimoncelliMask(img.shape[-2:], target=img, mask=mask)
            model.to(DEVICE2).to(torch.float64)
            met = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            met.setup(
                (torch.rand_like(img) - 0.5) * 0.1 + img.mean(),
                optimizer_kwargs={"lr": 0.02, "amsgrad": True},
            )
            met.synthesize(
                max_iter=1000, change_scale_criterion=None, ctf_iters_to_check=7
            )
            met.save("uploaded_files/ps_mask.pt")
            met_up = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            met_up.load(
                ps_regression / "ps_mask.pt",
                tensor_equality_atol=1e-7,
                map_location=DEVICE2,
            )
            compare_metamers(met, met_up)

        @pytest.mark.filterwarnings("ignore:You will need to call setup:UserWarning")
        @pytest.mark.filterwarnings(
            "ignore:initial_image and image are different sizes:UserWarning"
        )
        @pytest.mark.parametrize(
            "fn",
            [
                ("fig15e", "fig14e"),
                ("fig14b", "fig4a"),
                ("fig15a", "fig15b"),
            ],
        )
        def test_ps_mixture(self, ps_images, fn, ps_regression):
            torch.use_deterministic_algorithms(True)
            po.tools.set_seed(0)
            torch.save(
                torch.random.get_rng_state(),
                f"uploaded_files/torch_rng_state_ps_mixture_{'-'.join(fn)}.pt",
            )
            print(np.random.get_state())
            img = torch.cat(
                [
                    self.get_specific_img(*ps_images, fn[0]),
                    self.get_specific_img(*ps_images, fn[1]),
                ]
            ).to(DEVICE2)
            model = PortillaSimoncelliMixture(img.shape[-2:])
            model.to(DEVICE2).to(torch.float64)
            met = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            met.setup(
                (torch.rand_like(img[:1]) - 0.5) * 0.1 + img.mean(),
                optimizer_kwargs={"lr": 0.02, "amsgrad": True},
            )
            met.synthesize(
                max_iter=4000, change_scale_criterion=None, ctf_iters_to_check=7
            )
            met.save(f"uploaded_files/ps_mixture_{'-'.join(fn)}.pt")
            met_up = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            met_up.load(
                ps_regression / f"ps_mixture_{'-'.join(fn)}.pt",
                tensor_equality_atol=1e-7,
                map_location=DEVICE2,
            )
            compare_metamers(met, met_up)

        @pytest.mark.parametrize("mag_bool", [True, False])
        def test_ps_mag_means(self, ps_images, mag_bool, ps_regression):
            torch.use_deterministic_algorithms(True)
            po.tools.set_seed(100)
            torch.save(
                torch.random.get_rng_state(),
                f"uploaded_files/torch_rng_state_ps_mag_means-{mag_bool}.pt",
            )
            print(np.random.get_state())
            img = self.get_specific_img(*ps_images, "fig4a").to(DEVICE2)
            if mag_bool:
                model = PortillaSimoncelliMagMeans(img.shape[-2:])
            else:
                model = po.simul.PortillaSimoncelli(
                    img.shape[-2:], spatial_corr_width=7
                )
            model.to(DEVICE2).to(torch.float64)
            met = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            met.setup((torch.rand_like(img) - 0.5) * 0.1 + img.mean())
            met.synthesize(
                max_iter=3000,
                change_scale_criterion=None,
                ctf_iters_to_check=7,
            )
            met.save(f"uploaded_files/ps_mag_means-{mag_bool}.pt")
            met_up = po.synth.MetamerCTF(
                img,
                model,
                loss_function=po.tools.optim.l2_norm,
                coarse_to_fine="together",
            )
            met_up.load(
                ps_regression / f"ps_mag_means-{mag_bool}.pt",
                tensor_equality_atol=1e-7,
                map_location=DEVICE2,
            )
            compare_metamers(met, met_up)



"""
Functions that check for optimization convergence/stabilization.

The functions herein generally differ in what they are checking for
convergence: loss, pixel change, etc.

They should probably be able to accept the following arguments, in this order
(they can accept more):

- ``synth``: an OptimizedSynthesis object to check.

- ``stop_criterion``: the value used as criterion / tolerance that our
  convergence target is compared against.

- ``stop_iters_to_check``: how many iterations back to check for convergence.

They must return a single ``bool``: ``True`` if we've reached convergence,
``False`` if not.
"""

# to avoid circular import error:
# https://adamj.eu/tech/2021/05/13/python-type-hints-how-to-fix-circular-imports/
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..synthesize.metamer import MetamerCTF
    from ..synthesize.synthesis import OptimizedSynthesis


def loss_convergence(
    synth: "OptimizedSynthesis",
    stop_criterion: float,
    stop_iters_to_check: int,
) -> bool:
    r"""
    Check whether the loss has stabilized and, if so, return ``True``.

    We check whether:

    - We have been synthesizing for ``stop_iters_to_check`` iterations,
      i.e. ``len(synth.losses) > stop_iters_to_check``.

    - Loss has decreased by less than ``stop_criterion`` over the past
      ``stop_iters_to_check`` iterations.

    If both conditions are met, we return ``True``. Else, we return ``False``.

    Parameters
    ----------
    synth
        The OptimizedSynthesis object to check.
    stop_criterion
        If the loss over the past ``stop_iters_to_check`` has changed
        less than ``stop_criterion``, we terminate synthesis.
    stop_iters_to_check
        How many iterations back to check in order to see if the
        loss has stopped decreasing (for ``stop_criterion``).

    Returns
    -------
    loss_stabilized :
        Whether the loss has stabilized or not.
    """
    try:
        # get the private losses if available.
        losses = synth._losses
    except AttributeError:
        losses = synth.losses
    return (
        len(losses) > stop_iters_to_check
        and abs(losses[-stop_iters_to_check] - losses[-1]) < stop_criterion
    )


def coarse_to_fine_enough(synth: "MetamerCTF", i: int, ctf_iters_to_check: int) -> bool:
    r"""
    Check whether we've been synthesized all scales for long enough.

    This is meant to be paired with another convergence check, such as
    ``loss_convergence``.

    We check whether:

    - We have finished synthesizing each individual scale, i.e. ``synth.scales[0] ==
      "all"``.

    - We have been synthesizing all scales for more than ``ctf_iters_to_check``
      iterations, i.e. ``i - synth.scales_timing["all"][0]) > ctf_iters_to_check``.

    If both conditions are met, we return ``True``. Else, we return ``False``.

    Parameters
    ----------
    synth
        The MetamerCTF object to check.
    i
        The current iteration (0-indexed).
    ctf_iters_to_check
        Minimum number of iterations coarse-to-fine must run at each scale.
        If self.coarse_to_fine is ``False``, then this is ignored.

    Returns
    -------
    ctf_enough
        Whether we've been doing coarse to fine synthesis for long enough.
    """
    all_scales = synth.scales[0] == "all"
    # synth.scales_timing['all'] will only be a non-empty list if all_scales is
    # True, so we only check it then. This is equivalent to checking if both
    # conditions are true
    if all_scales:
        return (i - synth.scales_timing["all"][0]) > ctf_iters_to_check
    else:
        return False


def pixel_change_convergence(
    synth: "OptimizedSynthesis",
    stop_criterion: float,
    stop_iters_to_check: int,
) -> bool:
    """
    Check whether the pixel change norm has stabilized and, if so, return ``True``.

    We check whether:

    - We have been synthesizing for ``stop_iters_to_check`` iterations, i.e.
      ``len(synth.pixel_change_norm) > stop_iters_to_check``.

    - The ``pixel_change_norm`` has changed by less than ``stop_criterion`` over the
      past ``stop_iters_to_check`` iterations.

    If both conditions are met, we return ``True``. Else, we return ``False``.

    Parameters
    ----------
    synth
        The OptimizedSynthesis object to check.
    stop_criterion
        If the pixel change norm has been less than ``stop_criterion`` for all
        of the past ``stop_iters_to_check``, we terminate synthesis.
    stop_iters_to_check
        How many iterations back to check in order to see if the
        pixel change norm has stopped decreasing (for ``stop_criterion``).

    Returns
    -------
    loss_stabilized :
        Whether the pixel change norm has stabilized or not.
    """
    return (
        len(synth.pixel_change_norm) > stop_iters_to_check
        and (synth.pixel_change_norm[-stop_iters_to_check:] < stop_criterion).all()
    )



import os

import numpy as np
import pytest
import scipy.io as sio
import torch

import plenoptic as po
from conftest import DEVICE, IMG_DIR
from plenoptic.data.fetch import fetch_data


@pytest.fixture()
def test_files_dir():
    return fetch_data("plenoptic-test-files.tar.gz")


def test_find_files(test_files_dir):
    assert os.path.exists(os.path.join(test_files_dir, "buildSCFpyr0.mat"))


@pytest.fixture()
def ssim_images():
    return fetch_data("ssim_images.tar.gz")


@pytest.fixture()
def msssim_images():
    return fetch_data("msssim_images.tar.gz")


@pytest.fixture()
def ssim_analysis():
    ssim_analysis = fetch_data("ssim_analysis.mat")
    return sio.loadmat(ssim_analysis, squeeze_me=True)


@pytest.mark.parametrize(
    "paths",
    [
        IMG_DIR / "mixed",
        IMG_DIR / "256" / "einstein.pgm",
        [IMG_DIR / "256" / "einstein.pgm", IMG_DIR / "256" / "curie.pgm"],
    ],
)
@pytest.mark.parametrize("as_gray", [True, False])
def test_load_images(paths, as_gray):
    if paths == IMG_DIR / "mixed":
        # there are images of different sizes in here, which means we should raise
        # an Exception
        with pytest.raises(Exception):
            images = po.tools.data.load_images(paths, as_gray)
    else:
        images = po.tools.data.load_images(paths, as_gray)
        assert images.ndimension() == 4, "load_images did not return a 4d tensor!"


class TestPerceptualMetrics:
    @pytest.mark.parametrize("weighted", [True, False])
    def test_ssim_grad(self, einstein_img, curie_img, weighted):
        curie_img.requires_grad_()
        assert po.metric.ssim(einstein_img, curie_img, weighted=weighted).requires_grad
        curie_img.requires_grad_(False)

    def test_msssim_grad(self, einstein_img, curie_img):
        curie_img.requires_grad_()
        assert po.metric.ms_ssim(einstein_img, curie_img).requires_grad
        curie_img.requires_grad_(False)

    @pytest.mark.parametrize("func_name", ["ssim", "ms-ssim", "nlpd"])
    @pytest.mark.parametrize("size_A", [(), (3,), (1, 1), (6, 3), (6, 1), (6, 4)])
    @pytest.mark.parametrize("size_B", [(), (3,), (1, 1), (6, 3), (3, 1), (1, 4)])
    def test_batch_handling(self, einstein_img, curie_img, func_name, size_A, size_B):
        func = {
            "ssim": po.metric.ssim,
            "ms-ssim": po.metric.ms_ssim,
            "nlpd": po.metric.nlpd,
        }[func_name]
        A = einstein_img[0, 0].repeat(*size_A, 1, 1)
        B = curie_img[0, 0].repeat(*size_B, 1, 1)

        if not len(size_A) == len(size_B) == 2:
            with pytest.raises(
                Exception, match="Input images should have four dimensions"
            ):
                func(A, B)
        else:
            tgt_size = []
            for i in range(len(size_A)):
                if size_A[i] == size_B[i] or size_A[i] == 1 or size_B[i] == 1:
                    tgt_size.append(max(size_A[i], size_B[i]))
                else:
                    tgt_size = None
                    break
            if tgt_size is None:
                with pytest.raises(
                    ValueError,
                    match=(
                        "Either img1 and img2 should have the same number of"
                        " elements in the batch and channel dimensions, or "
                        "one of them should be 1"
                    ),
                ):
                    func(A, B)
            elif tgt_size[1] > 1:
                with pytest.warns(
                    Warning, match="computed separately for each channel"
                ):
                    assert func(A, B).shape == tuple(tgt_size)
            else:
                assert func(A, B).shape == tuple(tgt_size)

    @pytest.mark.parametrize("mode", ["many-to-one", "one-to-many"])
    def test_noise_independence(self, einstein_img, mode):
        # this makes sure that we are drawing the noise independently in the
        # two cases here
        if mode == "many-to-one":
            einstein_img = einstein_img.repeat(2, 1, 1, 1)
            noise_lvl = 1
        elif mode == "one-to-many":
            noise_lvl = [1, 1]
        noisy = po.tools.add_noise(einstein_img, noise_lvl)
        assert not torch.equal(*noisy)

    @pytest.mark.parametrize("noise_lvl", [[1], [128], [2, 4], [2, 4, 8], [0]])
    @pytest.mark.parametrize("noise_as_tensor", [True, False])
    def test_add_noise(self, einstein_img, noise_lvl, noise_as_tensor):
        if noise_as_tensor:
            noise_lvl = torch.as_tensor(
                noise_lvl, dtype=torch.float32, device=DEVICE
            ).unsqueeze(1)
        noisy = po.tools.add_noise(einstein_img, noise_lvl).to(DEVICE)
        if not noise_as_tensor:
            # always needs to be a tensor to properly check with allclose
            noise_lvl = torch.as_tensor(
                noise_lvl, dtype=torch.float32, device=DEVICE
            ).unsqueeze(1)
        assert torch.allclose(po.metric.mse(einstein_img, noisy), noise_lvl)

    @pytest.fixture
    def ssim_base_img(self, ssim_images, ssim_analysis):
        return po.load_images(os.path.join(ssim_images, ssim_analysis["base_img"])).to(
            DEVICE
        )

    @pytest.mark.parametrize("weighted", [True, False])
    @pytest.mark.parametrize("other_img", np.arange(1, 11))
    def test_ssim_analysis(
        self, weighted, other_img, ssim_images, ssim_analysis, ssim_base_img
    ):
        mat_type = {True: "weighted", False: "standard"}[weighted]
        other = po.load_images(os.path.join(ssim_images, f"samp{other_img}.tif")).to(
            DEVICE
        )
        # dynamic range is 1 for these images, because po.load_images
        # automatically re-ranges them. They were computed with
        # dynamic_range=255 in MATLAB, and by correctly setting this value,
        # that should be corrected for
        plen_val = po.metric.ssim(ssim_base_img, other, weighted)
        mat_val = torch.as_tensor(
            ssim_analysis[mat_type][f"samp{other_img}"].astype(np.float32),
            device=DEVICE,
        )
        # float32 precision is ~1e-6 (see `np.finfo(np.float32)`), and the
        # errors increase through multiplication and other operations.
        print(plen_val - mat_val, plen_val, mat_val)
        assert torch.allclose(plen_val, mat_val.view_as(plen_val), atol=1e-5)

    def test_msssim_analysis(self, msssim_images):
        # True values are defined by https://ece.uwaterloo.ca/~z70wang/research/iwssim/msssim.zip
        true_values = torch.as_tensor(
            [1.0000000, 0.9112161, 0.7699084, 0.8785111, 0.9488805],
            device=DEVICE,
        )
        computed_values = torch.zeros_like(true_values)
        base_img = po.load_images(os.path.join(msssim_images, "samp0.tiff")).to(DEVICE)
        for i in range(len(true_values)):
            other_img = po.load_images(os.path.join(msssim_images, f"samp{i}.tiff")).to(
                DEVICE
            )
            computed_values[i] = po.metric.ms_ssim(base_img, other_img)
        assert torch.allclose(true_values, computed_values)

    def test_nlpd_grad(self, einstein_img, curie_img):
        curie_img.requires_grad_()
        assert po.metric.nlpd(einstein_img, curie_img).requires_grad
        curie_img.requires_grad_(False)  # return to previous state for pytest fixtures

    @pytest.mark.parametrize("model", ["frontend.OnOff"], indirect=True)
    def test_model_metric_grad(self, einstein_img, curie_img, model):
        curie_img.requires_grad_()
        assert po.metric.model_metric(einstein_img, curie_img, model).requires_grad
        curie_img.requires_grad_(False)

    def test_ssim_dtype(self, einstein_img, curie_img):
        po.metric.ssim(einstein_img.to(torch.float64), curie_img.to(torch.float64))

    def test_ssim_dtype_exception(self, einstein_img, curie_img):
        with pytest.raises(ValueError, match="must have same dtype"):
            po.metric.ssim(einstein_img.to(torch.float64), curie_img)

    def test_msssim_dtype(self, einstein_img, curie_img):
        po.metric.ms_ssim(einstein_img.to(torch.float64), curie_img.to(torch.float64))

    def test_msssim_dtype_exception(self, einstein_img, curie_img):
        with pytest.raises(ValueError, match="must have same dtype"):
            po.metric.ms_ssim(einstein_img.to(torch.float64), curie_img)



#!/usr/bin/env python

import pathlib
import re
import sys

objects_to_check = []
try:
    import sphobjinv as soi

    inv = soi.Inventory("docs/_build/html/objects.inv")
    objects = [d.name for d in inv.objects if d.name.startswith("plenoptic.")]
    objects_to_check = []
    for obj in objects:
        # this will give a list of strings like (for
        # plenoptic.tools.validate.validate_model):
        # plenoptic.tools.validate.validate_model,
        # tools.validate.validate_model, validate.validate_model, validate_model
        obj = [".".join(obj.split(".")[i:]) for i in range(obj.count(".") + 1)]
        objects_to_check.extend(obj)
    objects_to_check = set(objects_to_check)
except ModuleNotFoundError:
    print("sphobjinv package not found, will not check for missing cross-references")
except TypeError:
    print(
        "sphinx objects.inv file not found, will not check for missing cross-references"
    )


paths = []
for p in sys.argv[1:]:
    p = pathlib.Path(p)
    if p.is_dir():
        p = list(p.glob("**/*.md"))
    elif p.suffix == ".md":
        p = [p]
    else:
        p = []
    paths.extend(p)

xrefs = []
missing_xrefs = []
double_backticks = []
method_format = []
for p in paths:
    with open(p) as f:
        md = f.read()
    if xr := re.findall(r"\[`.*?`\]\(?.*?\)?", md):
        xrefs.append((p, xr))
    miss_xr = []
    for obj in objects_to_check:
        if xr := re.findall(
            rf"`{obj}\(?\)?`(?! keyword)(?! argument)(?! ?<\!-- *?skip-lint *?-->)", md
        ):
            miss_xr.append(xr)
    if miss_xr:
        missing_xrefs.append((p, miss_xr))
    mfmt = []
    # these two strings are basically the same, but one is looking for "starts
    # with ." and the second is "followed by parentheses"
    for test in [
        r"`\.[A-Za-z0-9_]+\(?\)?(?: <.*?>|`)(?! ?<\!-- *?skip-lint *?-->)",
        r"`?[A-Za-z0-9_]+\(\)(?: <.*?>|`)",
    ]:
        mfmt.extend(re.findall(test, md))
    if mfmt:
        method_format.append((p, mfmt))
    if bt := re.findall(r"(?<!`)``(?!`).*(?<!`)``(?!`)", md):
        double_backticks.append((p, bt))

if xrefs or missing_xrefs or double_backticks or method_format:
    if xrefs:
        print("The following markdown documents' cross-references are misformatted!")
        for p, xr in xrefs:
            print(f"{p}:")
            for x in xr:
                print(f"\t{x}")
    if missing_xrefs:
        print("The following markdown documents' is missing some cross-references!")
        for p, xr in missing_xrefs:
            print(f"{p}:")
            for x in xr:
                print(f"\t{x}")
    if double_backticks:
        print("The following markdown documents contain double backticks!")
        for p, bt in double_backticks:
            print(f"{p}: {bt}")
    if method_format:
        print(
            "The following markdown documents contain method/attribute references"
            " that contain a . or (), remove them"
        )
        for p, mf in method_format:
            print(f"{p}: {mf}")
    sys.exit(1)
